import json
import re

from collections.abc import Iterator
from textwrap import dedent
from typing import TYPE_CHECKING, Any, Optional, Union

import yaml

from pydantic import StrictInt, StrictStr

from snowflake.core import PollingOperation
from snowflake.core._operation import PollingOperations

from .._common import CreateMode, SchemaObjectCollectionParent, SchemaObjectReferenceMixin
from .._generated.api_client import StoredProcApiClient
from .._internal.telemetry import api_telemetry
from .._internal.utils import deprecated


if TYPE_CHECKING:
    from snowflake.core.schema import SchemaResource

from snowflake.core.service._generated import (
    FetchServiceLogs200Response,
    FetchServiceStatus200Response,
    GrantOf,
    ServiceApi,
    ServiceContainer,
    ServiceEndpoint,
    ServiceInstance,
    ServiceRole,
    ServiceRoleGrantTo,
)
from snowflake.core.service._generated.models import JobService, Service, ServiceSpecInlineText, ServiceSpecStageFile


class ServiceCollection(SchemaObjectCollectionParent["ServiceResource"]):
    """Represents the collection operations on the Snowflake Snowpark Container Service resource.

    With this collection, you can create, update, iterate through, and fetch SPCSs that you have access to in the
    current context.

    Examples
    ________
    Creating a service instance:

    >>> my_service = Service(
    ...     name="my_service",
    ...     min_instances=1,
    ...     max_instances=2,
    ...     compute_pool="my_compute_pool",
    ...     spec=ServiceSpec("@my_stage/my_service_spec.yaml"),
    ... )
    >>> root.databases["my_db"].schemas["my_schema"].services.create(my_service)
    """

    def __init__(self, schema: "SchemaResource"):
        super().__init__(schema, ServiceResource)
        self._api = ServiceApi(
            root=self.root, resource_class=self._ref_class, sproc_client=StoredProcApiClient(root=self.root)
        )

    @api_telemetry
    def iter(
        self,
        *,
        like: Optional[str] = None,
        starts_with: Optional[str] = None,
        limit: Optional[int] = None,
        from_name: Optional[str] = None,
    ) -> Iterator[Service]:
        """Iterate through ``Service`` objects from Snowflake, filtering on any optional 'like' pattern.

        Parameters
        __________
        like: str, optional
            A case-insensitive string functioning as a filter, with support for SQL
            wildcard characters (% and _).
        starts_with: str, optional
            String used to filter the command output based on the string of characters that appear
            at the beginning of the object name. Uses case-sensitive pattern matching.
        show_limit: int, optional
            Limit of the maximum number of rows returned by iter(). The default is ``None``, which behaves equivalently
            to show_limit=10000. This value must be between ``1`` and ``10000``.
        from_name: str, optional
            Fetch rows only following the first row whose object name matches
            the specified string. This is case-sensitive and does not have to be the full name.

        Examples
        ________
        Showing all services that you have access to see:

        >>> services = service_collection.iter()

        Showing information of the exact service you want to see:

        >>> services = service_collection.iter(like="your-service-name")

        Showing services starting with 'your-service-name-':

        >>> services = service_collection.iter(like="your-service-name-%")

        Using a for loop to retrieve information from iterator:

        >>> for service in services:
        >>>    print(service.name)
        """
        services = self._api.list_services(
            self.database.name,
            self.schema.name,
            StrictStr(like) if like is not None else None,
            StrictStr(starts_with) if starts_with else None,
            limit,
            from_name,
            async_req=False,
        )

        return iter(services)

    @api_telemetry
    def iter_async(
        self,
        *,
        like: Optional[str] = None,
        starts_with: Optional[str] = None,
        limit: Optional[int] = None,
        from_name: Optional[str] = None,
    ) -> PollingOperation[Iterator[Service]]:
        """An asynchronous version of :func:`iter`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self._api.list_services(
            self.database.name,
            self.schema.name,
            StrictStr(like) if like is not None else None,
            StrictStr(starts_with) if starts_with else None,
            limit,
            from_name,
            async_req=True,
        )
        return PollingOperations.iterator(future)

    @api_telemetry
    def create(self, service: Service, *, mode: CreateMode = CreateMode.error_if_exists) -> "ServiceResource":
        """Create a Snowpark Container service in Snowflake.

        Parameters
        __________
        service: Service
            The ``Service`` object, together with the ``Service``'s properties:
            name, compute_pool, spec; auto_resume, min_instances, max_instances, status, external_access_integrations,
            query_warehouse, comment are optional.
        mode: CreateMode, optional
        One of the following enum values.

        ``CreateMode.error_if_exists``: Throw an :class:`snowflake.core.exceptions.ConflictError`
        if the service already exists in Snowflake.  Equivalent to SQL ``create service <name> ...``.

        ``CreateMode.or_replace``: Replace if the service already exists in Snowflake. Equivalent to SQL
        ``create or replace service <name> ...``.

        ``CreateMode.if_not_exists``: Do nothing if the service already exists in Snowflake.
        Equivalent to SQL ``create service <name> if not exists...``

        Default is ``CreateMode.error_if_exists``.

        Examples
        ________
        Creating a service, replacing any existing service with the same name:

        >>> services = root.databases["my_db"].schemas["my_schema"].services
        >>> my_service = Service(
        ...     name="my_service",
        ...     compute_pool="my_compute_pool",
        ...     spec=ServiceSpec("@my_stage/my_service_spec.yaml"),
        ... )
        >>> services.create(my_service, mode=CreateMode.or_replace)
        """
        if mode == CreateMode.or_replace:
            raise ValueError(f"{mode} is not a valid value for this resource")
        real_mode = CreateMode[mode].value
        self._api.create_service(self.database.name, self.schema.name, service, StrictStr(real_mode), async_req=False)
        return self[service.name]

    @api_telemetry
    def create_async(
        self, service: Service, *, mode: CreateMode = CreateMode.error_if_exists
    ) -> PollingOperation["ServiceResource"]:
        """An asynchronous version of :func:`create`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        if mode == CreateMode.or_replace:
            raise ValueError(f"{mode} is not a valid value for this resource")
        real_mode = CreateMode[mode].value
        future = self._api.create_service(
            self.database.name, self.schema.name, service, StrictStr(real_mode), async_req=True
        )
        return PollingOperation(future, lambda _: self[service.name])

    @api_telemetry
    def execute_job(self, job_service: JobService) -> "ServiceResource":
        """Execute a Snowpark Container job service in Snowflake.

        Parameters
        __________
        job_service: JobService
            The ``JobService`` object, together with the ``JobService``'s properties:
            name, compute_pool, spec; status, external_access_integrations, query_warehouse, comment are optional

        Examples
        ________
        Executing a job service:

        >>> job_service = JobService(
        ...     name="my_job_service", compute_pool="my_cp", spec=ServiceSpec("@my_stage/my_service_spec.yaml")
        ... )
        >>> services.execute_job(job_service)
        """
        self._api.execute_job_service(self.database.name, self.schema.name, job_service, async_req=False)
        return self[job_service.name]

    @api_telemetry
    def execute_job_async(self, job_service: JobService) -> PollingOperation["ServiceResource"]:
        """An asynchronous version of :func:`execute_job`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self._api.execute_job_service(self.database.name, self.schema.name, job_service, async_req=True)
        return PollingOperation(future, lambda _: self[job_service.name])


class ServiceResource(SchemaObjectReferenceMixin[ServiceCollection]):
    """Represents a reference to a Snowflake Snowpark Container Service.

    With this service reference, you can fetch information about a service, as well as perform
    certain actions on it.
    """

    def __init__(self, name: str, collection: ServiceCollection) -> None:
        self.name = name
        self.collection = collection

    @api_telemetry
    def create_or_alter(self, service: Service) -> None:
        """Create a service in Snowflake or alter one if it already exists.

        Parameters
        __________
        service: Service
            The ``Service`` object, together with the ``Service``'s properties:
            name, compute_pool, spec; auto_resume, min_instances, max_instances, status, external_access_integrations,
            query_warehouse, comment are optional.

        Examples
        ________
        Creating or updating a service in Snowflake:

        >>> service_parameters = Service(
        ...     name="your-service-name",
        ...     compute_pool="my_cp"
        ...     spec=ServiceSpecStageFile(stage="stage_name", spec_file=spec_file),
        ...)
        >>> services = root.databases["my_db"].schemas["my_schema"].services
        >>> services["your-service-name"].create_or_alter(service_parameters)
        """
        self.collection._api.create_or_alter_service(
            self.database.name, self.schema.name, self.name, service, async_req=False
        )

    @api_telemetry
    def create_or_alter_async(self, service: Service) -> PollingOperation[None]:
        """An asynchronous version of :func:`create_or_alter`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.create_or_alter_service(
            self.database.name, self.schema.name, self.name, service, async_req=True
        )
        return PollingOperations.empty(future)

    @api_telemetry
    @deprecated("drop")
    def delete(self) -> None:
        self.drop()

    @api_telemetry
    def drop(self, if_exists: Optional[bool] = None) -> None:
        """Drop the service.

        Parameters
        __________
        if_exists: bool, optional
            Check the existence of this service before dropping it.
            Default is ``None``, which is equivalent to ``False``.

        Examples
        ________
        Deleting a service using its reference:

        >>> service_reference.drop()
        """
        self.collection._api.delete_service(self.database.name, self.schema.name, self.name, if_exists, async_req=False)

    @api_telemetry
    def drop_async(self, if_exists: Optional[bool] = None) -> PollingOperation[None]:
        """An asynchronous version of :func:`drop`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.delete_service(
            self.database.name, self.schema.name, self.name, if_exists, async_req=True
        )
        return PollingOperations.empty(future)

    @api_telemetry
    def fetch(self) -> Service:
        """Fetch the details of a service.

        Examples
        ________
        Fetching a service using its reference:

        >>> service_reference.fetch()
        """
        return self.collection._api.fetch_service(self.database.name, self.schema.name, self.name, async_req=False)

    @api_telemetry
    def fetch_async(self) -> PollingOperation[Service]:
        """An asynchronous version of :func:`fetch`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.fetch_service(self.database.name, self.schema.name, self.name, async_req=True)
        return PollingOperations.identity(future)

    @api_telemetry
    def suspend(self, if_exists: Optional[bool] = None) -> None:
        """Suspend the service.

        Parameters
        __________
        if_exists: bool, optional
            Check the existence of this service before suspending it.
            Default is ``None``, which is equivalent to ``False``.

        Examples
        ________
        Suspending a service using its reference:

        >>> service_reference.suspend()
        """
        self.collection._api.suspend_service(
            self.database.name, self.schema.name, self.name, if_exists, async_req=False
        )

    @api_telemetry
    def suspend_async(self, if_exists: Optional[bool] = None) -> PollingOperation[None]:
        """An asynchronous version of :func:`suspend`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.suspend_service(
            self.database.name, self.schema.name, self.name, if_exists, async_req=True
        )
        return PollingOperations.empty(future)

    @api_telemetry
    def resume(self, if_exists: Optional[bool] = None) -> None:
        """Resumes the service.

        Parameters
        __________
        if_exists: bool, optional
            Check the existence of this service before resuming it.
            Default is ``None``, which is equivalent to ``False``.

        Examples
        ________
        Resuming a service using its reference:

        >>> service_reference.resume()
        """
        self.collection._api.resume_service(self.database.name, self.schema.name, self.name, if_exists, async_req=False)

    @api_telemetry
    def resume_async(self, if_exists: Optional[bool] = None) -> PollingOperation[None]:
        """An asynchronous version of :func:`resume`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.resume_service(
            self.database.name, self.schema.name, self.name, if_exists, async_req=True
        )
        return PollingOperations.empty(future)

    @api_telemetry
    def get_endpoints(self) -> Iterator[ServiceEndpoint]:
        """Show the endpoints corresponding to this service.

        Examples
        ________
        Showing the endpoints of a service using its reference:

        >>> service_reference.get_endpoints()
        """
        return iter(
            self.collection._api.show_service_endpoints(
                self.database.name, self.schema.name, self.name, async_req=False
            )
        )

    @api_telemetry
    def get_endpoints_async(self) -> PollingOperation[Iterator[ServiceEndpoint]]:
        """An asynchronous version of :func:`get_endpoints`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.show_service_endpoints(
            self.database.name, self.schema.name, self.name, async_req=True
        )
        return PollingOperations.iterator(future)

    @api_telemetry
    def get_containers(self) -> Iterator[ServiceContainer]:
        """Show the containers corresponding to this service.

        Examples
        ________
        Showing the containers of a service using its reference:

        >>> service_reference.get_containers()
        """
        return iter(
            self.collection._api.list_service_containers(
                self.database.name, self.schema.name, self.name, async_req=False
            )
        )

    @api_telemetry
    def get_containers_async(self) -> PollingOperation[Iterator[ServiceContainer]]:
        """An asynchronous version of :func:`get_containers`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.list_service_containers(
            self.database.name, self.schema.name, self.name, async_req=True
        )
        return PollingOperations.iterator(future)

    @api_telemetry
    def get_instances(self) -> Iterator[ServiceInstance]:
        """Show the instances corresponding to this service.

        Examples
        ________
        Showing the instances of a service using its reference:

        >>> service_reference.get_instances()
        """
        return iter(
            self.collection._api.list_service_instances(
                self.database.name, self.schema.name, self.name, async_req=False
            )
        )

    @api_telemetry
    def get_instances_async(self) -> PollingOperation[Iterator[ServiceInstance]]:
        """An asynchronous version of :func:`get_instances`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.list_service_instances(
            self.database.name, self.schema.name, self.name, async_req=True
        )
        return PollingOperations.iterator(future)

    @api_telemetry
    def get_roles(self) -> Iterator[ServiceRole]:
        """Show the roles corresponding to this service.

        Examples
        ________
        Showing the roles of a service using its reference:

        >>> service_reference.get_roles()
        """
        return iter(
            self.collection._api.list_service_roles(self.database.name, self.schema.name, self.name, async_req=False)
        )

    @api_telemetry
    def get_roles_async(self) -> PollingOperation[Iterator[ServiceRole]]:
        """An asynchronous version of :func:`get_roles`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.list_service_roles(
            self.database.name, self.schema.name, self.name, async_req=True
        )
        return PollingOperations.iterator(future)

    @api_telemetry
    def iter_grants_of_service_role(self, role_name: str) -> Iterator[GrantOf]:
        """Show the grants of the service role associated with this service.

        Parameters
        __________
        role_name: str
            Name of the service role.

        Examples
        ________
        Showing the grants of the service role associated with a service using the service reference:

        >>> service_reference.iter_grants_of_service_role("all_endpoints_usage")
        """
        return iter(
            self.collection._api.list_service_role_grants_of(
                self.database.name, self.schema.name, self.name, role_name, async_req=False
            )
        )

    @api_telemetry
    def iter_grants_of_service_role_async(self, role_name: str) -> PollingOperation[Iterator[GrantOf]]:
        """An asynchronous version of :func:`iter_grants_of_service_role`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.list_service_role_grants_of(
            self.database.name, self.schema.name, self.name, role_name, async_req=True
        )
        return PollingOperations.iterator(future)

    @api_telemetry
    def iter_grants_to_service_role(self, role_name: str) -> Iterator[ServiceRoleGrantTo]:
        """Show the grants given to the service role associated with this service.

        Parameters
        __________
        role_name: str
            Name of the service role.

        Examples
        ________
        Showing the grants given to the service role associated with a service using the service reference:

        >>> service_reference.iter_grants_to_service_role("all_endpoints_usage")
        """
        return iter(
            self.collection._api.list_service_role_grants_to(
                self.database.name, self.schema.name, self.name, role_name, async_req=False
            )
        )

    @api_telemetry
    def iter_grants_to_service_role_async(self, role_name: str) -> PollingOperation[Iterator[ServiceRoleGrantTo]]:
        """An asynchronous version of :func:`iter_grants_to_service_role`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.list_service_role_grants_to(
            self.database.name, self.schema.name, self.name, role_name, async_req=True
        )
        return PollingOperations.iterator(future)

    @api_telemetry
    @deprecated("get_containers")
    def get_service_status(self, timeout: int = 0) -> list[dict[str, Any]]:
        """Get the status of the service.

        Parameters
        __________
        timeout: int
          Number of seconds to wait for the service to reach a steady state (for example, READY)
          before returning the status. If the service does not reach steady state within the specified time,
          Snowflake returns the current state.

          If not specified or ``0``, Snowflake returns the current state immediately.

          Default: ``0`` seconds.

        Examples
        ________
        Getting the status of a service using its reference:

        >>> service_reference.get_service_status()

        Getting the status of a service using its reference with a timeout:

        >>> service_reference.get_service_status(timeout=10)
        """
        status = self.collection._api.fetch_service_status(
            self.database.name, self.schema.name, self.name, StrictInt(timeout), async_req=False
        )
        if status.systemget_service_status is None:
            return list()
        return json.loads(status.systemget_service_status)

    @api_telemetry
    @deprecated("get_containers_async")
    def get_service_status_async(self, timeout: int = 0) -> PollingOperation[list[dict[str, Any]]]:
        """An asynchronous version of :func:`get_service_status`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.fetch_service_status(
            self.database.name, self.schema.name, self.name, StrictInt(timeout), async_req=True
        )

        def transform(status: FetchServiceStatus200Response) -> list[dict[str, Any]]:
            if status.systemget_service_status is None:
                return list()
            return json.loads(status.systemget_service_status)

        return PollingOperation(future, transform)

    @api_telemetry
    def get_service_logs(self, instance_id: str, container_name: str, num_lines: Optional[int] = None) -> str:
        """Get the service logs of the service.

        Parameters
        __________
        instance_id: str
            Instance ID of the service.
        container_name: str
            Container name of the service.
        num_lines: int, optional
            Number of the most recent log lines to retrieve.


        :meth:`get_service_status` returns the ``instance_id`` and ``container_name`` as a part of its results.

        Examples
        ________
        Getting the logs of a service using its reference:

        >>> service_reference.get_service_logs(instance_id="instance_id", container_name="container_name")
        """
        logs = self.collection._api.fetch_service_logs(
            self.database.name,
            self.schema.name,
            self.name,
            StrictInt(instance_id),
            StrictStr(container_name),
            num_lines,
            async_req=False,
        )
        if logs.systemget_service_logs is None:
            return ""
        return logs.systemget_service_logs

    @api_telemetry
    def get_service_logs_async(
        self, instance_id: str, container_name: str, num_lines: Optional[int] = None
    ) -> PollingOperation[str]:
        """An asynchronous version of :func:`get_service_logs`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self.collection._api.fetch_service_logs(
            self.database.name,
            self.schema.name,
            self.name,
            StrictInt(instance_id),
            StrictStr(container_name),
            num_lines,
            async_req=True,
        )

        def transform(logs: FetchServiceLogs200Response) -> str:
            if logs.systemget_service_logs is None:
                return ""
            return logs.systemget_service_logs

        return PollingOperation(future, transform)


def ServiceSpec(spec: str) -> Union[ServiceSpecInlineText, ServiceSpecStageFile]:
    """Infers whether a specification is a stage file or inline text.

    Any spec that starts with '@' is parsed as a stage file, otherwise it is passed as an inline text.
    """
    spec = dedent(spec).rstrip()
    if spec.startswith("@"):
        stage, spec_file = _parse_spec_path(spec[1:])
        return ServiceSpecStageFile(stage=stage, spec_file=spec_file)
    else:
        if _validate_inline_spec(spec):
            return ServiceSpecInlineText(spec_text=spec)
        else:
            raise ValueError(f"{spec} is not a valid service spec inline text")


def _validate_inline_spec(spec_str: str) -> bool:
    try:
        spec_data = yaml.safe_load(spec_str)
    except yaml.YAMLError:
        return False
    if not isinstance(spec_data, dict) or "spec" not in spec_data:
        return False
    return True


def _parse_spec_path(spec_path: str) -> tuple[str, str]:
    # this pattern tries to match a file path depth of at least two (needs a stage and file name at minimum)
    pattern = r'^[^<>:"|?*\/\n]+(?:\/[^<>:"|?*\/\n]+)+$'
    if not re.fullmatch(pattern, spec_path):
        raise ValueError(f"{spec_path} is not a valid stage file path")
    stage, path = spec_path.split("/", 1)
    return stage, path



"""
Cortex Agent API.

OpenAPI 3.0 specification for the Cortex Agent API  # noqa: E501

The version of the OpenAPI document: 0.1.0
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, Dict, List, Optional

from pydantic import BaseModel, ConfigDict, StrictStr


class ToolToolSpecInputSchema(BaseModel):
    """A model object representing the ToolToolSpecInputSchema resource.

    Constructs an object of type ToolToolSpecInputSchema with the provided properties.

    Parameters
    __________
    type : str, optional
        The type of the input schema object
    properties : object, optional
        Definitions of each input parameter
    required : list[str], optional
        List of required input parameter names
    """

    type: Optional[StrictStr] = None

    properties: Optional[Dict[str, Any]] = None

    required: Optional[List[StrictStr]] = None

    additional_properties: dict[str, Any] = {}

    __properties = ["type", "properties", "required"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> ToolToolSpecInputSchema:
        """Create an instance of ToolToolSpecInputSchema from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = {"additional_properties"}

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        # puts key-value pairs in additional_properties in the top level
        if self.additional_properties is not None:
            for _key, _value in self.additional_properties.items():
                _dict[_key] = _value

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> ToolToolSpecInputSchema:
        """Create an instance of ToolToolSpecInputSchema from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return ToolToolSpecInputSchema.model_validate(obj)

        _obj = ToolToolSpecInputSchema.model_validate(
            {
                "type": obj.get("type"),
                "properties": obj.get("properties"),
                "required": obj.get("required"),
            }
        )

        # store additional fields in additional_properties
        for _key in obj.keys():
            if _key not in cls.__properties:
                _obj.additional_properties[_key] = obj.get(_key)

        return _obj


class ToolToolSpecInputSchemaModel:
    def __init__(
        self,  # optional properties
        type: Optional[str] = None,
        properties: Optional[object] = None,
        required: Optional[list[str]] = None,
    ):
        """A model object representing the ToolToolSpecInputSchema resource.

        Constructs an object of type ToolToolSpecInputSchema with the provided properties.

        Parameters
        __________
        type : str, optional
            The type of the input schema object
        properties : object, optional
            Definitions of each input parameter
        required : list[str], optional
            List of required input parameter names
        """
        self.type = type
        self.properties = properties
        self.required = required

    additional_properties: dict[str, Any] = {}

    __properties = ["type", "properties", "required"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return ToolToolSpecInputSchema(
            type=self.type,
            properties=self.properties,
            required=self.required,
        )

    @classmethod
    def _from_model(cls, model) -> ToolToolSpecInputSchemaModel:
        return ToolToolSpecInputSchemaModel(
            type=model.type,
            properties=model.properties,
            required=model.required,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a ToolToolSpecInputSchema.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the ToolToolSpecInputSchema object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> ToolToolSpecInputSchemaModel:
        """Create an instance of ToolToolSpecInputSchema from a dict.

        This method constructs a ToolToolSpecInputSchema object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        ToolToolSpecInputSchema
            A ToolToolSpecInputSchema object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(ToolToolSpecInputSchema.from_dict(obj))


ToolToolSpecInputSchema._model_class = ToolToolSpecInputSchemaModel



from __future__ import annotations

import re

from typing import TYPE_CHECKING, Any

from snowflake.core.exceptions import InvalidIdentifierError


if TYPE_CHECKING:
    from snowflake.connector import SnowflakeConnection


# See https://docs.snowflake.com/en/sql-reference/identifiers-syntax for identifier syntax
UNQUOTED_IDENTIFIER_REGEX = r"([a-zA-Z_])([a-zA-Z0-9_$]{0,254})"
QUOTED_IDENTIFIER_REGEX = r'"((""|[^"]){0,255})"'
VALID_IDENTIFIER_REGEX = f"(?:{UNQUOTED_IDENTIFIER_REGEX}|{QUOTED_IDENTIFIER_REGEX})"


class FQN:
    """Represents an object identifier, supporting fully qualified names.

    The instance supports builder pattern that allows updating the identifier with database and
    schema from different sources.

    Examples
    ________
    >>> fqn = FQN.from_string("my_schema.object").using_connection(conn)

    >>> fqn = FQN.from_string("my_name").set_database("db").set_schema("foo")
    """

    def __init__(self, database: str | None, schema: str | None, name: str, signature: str | None = None) -> None:
        self._database = database
        self._schema = schema
        self._name = name
        self.signature = signature

    @property
    def database(self) -> str | None:
        return self._database

    @property
    def schema(self) -> str | None:
        return self._schema

    @property
    def name(self) -> str:
        return self._name

    @property
    def prefix(self) -> str:
        if self.database:
            return f"{self.database}.{self.schema if self.schema else 'PUBLIC'}"
        if self.schema:
            return f"{self.schema}"
        return ""

    @property
    def identifier(self) -> str:
        if self.prefix:
            return f"{self.prefix}.{self.name}"
        return self.name

    def __str__(self) -> str:
        return self.identifier

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, FQN):
            return NotImplemented
        return self.identifier == other.identifier

    @classmethod
    def from_string(cls, identifier: str) -> FQN:
        """Take in an object name in the form [[database.]schema.]name and return a new :class:`FQN` instance.

        Raises:
            InvalidIdentifierError: If the object identifier does not meet identifier requirements.
        """
        qualifier_pattern = (
            rf"(?:(?P<first_qualifier>{VALID_IDENTIFIER_REGEX})\.)?"
            rf"(?:(?P<second_qualifier>{VALID_IDENTIFIER_REGEX})\.)?"
            rf"(?P<name>{VALID_IDENTIFIER_REGEX})(?P<signature>\(.*\))?"
        )
        result = re.fullmatch(qualifier_pattern, identifier)

        if result is None:
            raise InvalidIdentifierError(identifier)

        unqualified_name = result.group("name")
        if result.group("second_qualifier") is not None:
            database = result.group("first_qualifier")
            schema = result.group("second_qualifier")
        else:
            database = None
            schema = result.group("first_qualifier")

        signature = None
        if result.group("signature"):
            signature = result.group("signature")
        return cls(name=unqualified_name, schema=schema, database=database, signature=signature)

    def set_database(self, database: str | None) -> FQN:
        if database:
            self._database = database
        return self

    def set_schema(self, schema: str | None) -> FQN:
        if schema:
            self._schema = schema
        return self

    def set_name(self, name: str) -> FQN:
        self._name = name
        return self

    def using_connection(self, conn: SnowflakeConnection) -> FQN:
        """Update the instance with database and schema from connection."""
        # Update the identifier only it if wasn't already a qualified name
        if conn.database and not self.database:
            self.set_database(conn.database)
        if conn.schema and not self.schema:
            self.set_schema(conn.schema)
        return self

    def to_dict(self) -> dict[str, str | None]:
        """Return the dictionary representation of the instance."""
        return {"name": self.name, "schema": self.schema, "database": self.database}



"""Manages Snowflake Pipes.

Example:
    >>> pipes: PipeCollection = root.databases["mydb"].schemas["myschema"].pipes
    >>> mypipe = pipes.create(Pipe("mypipe"))
    >>> pipe_iter = pipes.iter(like="my%")
    >>> pipe = pipes["mypipe"]
    >>> an_existing_pipe = pipes["an_existing_pipe"]

Refer to :class:`snowflake.core.Root` to create the ``root``.
"""

from ._generated.models import Pipe
from ._pipe import PipeCollection, PipeResource


__all__ = ["Pipe", "PipeCollection", "PipeResource"]



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#
import datetime

from io import BytesIO
from textwrap import dedent

import pytest

from snowflake.core.service import Service, ServiceSpecStageFile
from tests.utils import random_string


pytestmark = [pytest.mark.skip_gov]


@pytest.mark.usefixtures("qa_mode_enabled")
def test_iter(services, temp_service, session, imagerepo, shared_compute_pool, database, schema):
    service_name = random_string(5, "test_service_a_")
    stage_name = random_string(5, "test_stage_")
    session.sql(f"create temp stage {stage_name};").collect()
    spec_file = "spec.yaml"
    stage_file = f"@{stage_name}"
    spec = f"{stage_file}/{spec_file}"
    session.file.put_stream(
        BytesIO(
            dedent(f"""
                spec:
                  containers:
                  - name: hello-world
                    image: {imagerepo}/hello-world:latest
                """).encode()
        ),
        spec,
    )
    test_service = Service(
        name=service_name,
        compute_pool=shared_compute_pool,
        spec=ServiceSpecStageFile(stage=stage_name, spec_file=spec_file),
        min_instances=1,
        max_instances=1,
    )

    try:
        s = services.create(test_service)
        service_names = [sr.name for sr in services.iter()]
        assert temp_service.name.upper() in service_names

        service_names = [sr.name for sr in services.iter(like="TESt_Serv%")]
        assert temp_service.name.upper() in service_names
        assert service_name.upper() in service_names
        service_names = [sr.name for sr in services.iter(starts_with="TESt_Serv")]
        assert temp_service.name.upper() not in service_names
        assert service_name.upper() not in service_names

        service_names = [sr.name for sr in services.iter(starts_with="TEST_SERVICE")]
        assert temp_service.name.upper() in service_names
        assert service_name.upper() in service_names

        service_names = [sr.name for sr in services.iter(starts_with="test_service")]
        assert temp_service.name.upper() not in service_names
        assert service_name.upper() not in service_names

        service_names = [sr.name for sr in services.iter(like="test_service_%")]
        assert temp_service.name.upper() in service_names

        service_names = [sr.name for sr in services.iter(like="test_service_%")]
        assert len(service_names) >= 2
        assert temp_service.name.upper() in service_names
        assert service_name.upper() in service_names

        service_names = [sr.name for sr in services.iter(like="test_service_%", limit=1)]
        assert len(service_names) == 1

        service_names = [sr.name for sr in services.iter(from_name="TEST_SERVICE_%", limit=1)]
        assert len(service_names) == 1

        service_names = [sr.name for sr in services.iter(from_name="TEST_SERVICE_%")]
        assert len(service_names) >= 2

        # Verifying the output of the show service command for the service created in the test file
        result = list(services.iter(like=f"{service_name.upper()}"))
        assert len(result) == 1
        expected_spec_for_new_service = (
            '---\nspec:\n  containers:\n  - name: "hello-world"'
            f'\n    image: "{imagerepo}/hello-world:latest"\n    resources:\n      '
            'limits:\n        memory: "6Gi"\n        cpu: "8"\n      requests:\n        '
            'memory: "0.5Gi"\n        cpu: "0.5"\n'
        )
        assert_service_output(
            result[0],
            service_name,
            shared_compute_pool,
            expected_spec_for_new_service,
            database.name,
            schema.name,
            1,
            1,
        )
        assert s.fetch().to_dict() == result[0].to_dict()

        # Verifying the output of the show service command for the temp created by the test setup
        result = list(services.iter(like=f"{temp_service.name.upper()}"))
        assert len(result) == 1
        expected_spec_for_temp_service = (
            '---\nspec:\n  containers:\n  - name: "hello-world"'
            f'\n    image: "{imagerepo}/hello-world:latest"\n    resources:\n      '
            'limits:\n        memory: "6Gi"\n        cpu: "8"\n      requests:\n   '
            '     memory: "0.5Gi"\n        cpu: "0.5"\n  endpoints:\n  - name: '
            '"default"\n    port: 8080\n'
        )
        assert_service_output(
            result[0],
            temp_service.name,
            shared_compute_pool,
            expected_spec_for_temp_service,
            database.name,
            schema.name,
            1,
            5,
            "created by temp_service",
        )
        assert temp_service.fetch().to_dict() == result[0].to_dict()
    finally:
        s.drop()


def assert_service_output(
    service,
    expected_service_name,
    expected_compute_pool_name,
    expected_spec_text,
    expected_database_name,
    expected_schema_name,
    expected_min_instances,
    expected_max_instances,
    expected_comment=None,
):
    assert service.name == expected_service_name.upper()
    assert service.status in ["PENDING", "RUNNING"]
    assert service.compute_pool == expected_compute_pool_name.upper()
    assert service.spec is not None
    assert service.spec.spec_text == expected_spec_text
    assert service.external_access_integrations is None
    assert service.query_warehouse is None
    assert service.comment == expected_comment
    assert service.auto_resume is True
    assert service.current_instances == 1
    assert service.target_instances == 1
    assert service.min_instances == expected_min_instances
    assert service.max_instances == expected_max_instances
    assert service.database_name == expected_database_name.upper()
    assert service.schema_name == expected_schema_name.upper()
    assert service.owner == "ACCOUNTADMIN"
    assert service.dns_name is not None and f"{expected_service_name.lower().replace('_', '-')}" in service.dns_name
    assert service.created_on == datetime.datetime(1967, 6, 23, 7, 0, 0, 123000, tzinfo=datetime.timezone.utc)
    assert service.updated_on == datetime.datetime(1967, 6, 23, 7, 0, 0, 123000, tzinfo=datetime.timezone.utc)
    assert service.resumed_on == datetime.datetime(1967, 6, 23, 7, 0, 0, 123000, tzinfo=datetime.timezone.utc)
    assert service.owner_role_type == "ROLE"
    assert service.is_job is False
    assert service.spec_digest is not None
    assert service.is_upgrading is False
    assert service.managing_object_domain is None
    assert service.managing_object_name is None



"""
Snowflake Iceberg Table API.

The Snowflake Iceberg Table API is a REST API that you can use to access, update, and perform certain actions on Iceberg Table resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, StrictBool, StrictStr


class IcebergTableFromAWSGlueCatalog(BaseModel):
    """A model object representing the IcebergTableFromAWSGlueCatalog resource.

    Constructs an object of type IcebergTableFromAWSGlueCatalog with the provided properties.

    Parameters
    __________
    name : str
        Specifies the name for the table, must be unique for the schema in which the table is created
    catalog_table_name : str
        Specifies the table name as recognized by the AWS Glue Data Catalog.
    external_volume : str, optional
        Specifies the name of the external volume to use for the table
    catalog_namespace : str, optional
        Specifies the catalog namespace for the table.
    replace_invalid_characters : bool, optional
        Specifies whether to replace invalid characters in the column names
    auto_refresh : bool, optional
        Specifies whether to automatically refresh the table metadata
    catalog : str, optional
        Specifies the name of the catalog integration to use for the table
    comment : str, optional
        Specifies a comment for the table
    """

    name: StrictStr

    external_volume: Optional[StrictStr] = None

    catalog_table_name: StrictStr

    catalog_namespace: Optional[StrictStr] = None

    replace_invalid_characters: Optional[StrictBool] = None

    auto_refresh: Optional[StrictBool] = None

    catalog: Optional[StrictStr] = None

    comment: Optional[StrictStr] = None

    __properties = [
        "name",
        "external_volume",
        "catalog_table_name",
        "catalog_namespace",
        "replace_invalid_characters",
        "auto_refresh",
        "catalog",
        "comment",
    ]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> IcebergTableFromAWSGlueCatalog:
        """Create an instance of IcebergTableFromAWSGlueCatalog from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> IcebergTableFromAWSGlueCatalog:
        """Create an instance of IcebergTableFromAWSGlueCatalog from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return IcebergTableFromAWSGlueCatalog.model_validate(obj)

        _obj = IcebergTableFromAWSGlueCatalog.model_validate(
            {
                "name": obj.get("name"),
                "external_volume": obj.get("external_volume"),
                "catalog_table_name": obj.get("catalog_table_name"),
                "catalog_namespace": obj.get("catalog_namespace"),
                "replace_invalid_characters": obj.get("replace_invalid_characters"),
                "auto_refresh": obj.get("auto_refresh"),
                "catalog": obj.get("catalog"),
                "comment": obj.get("comment"),
            }
        )

        return _obj


class IcebergTableFromAWSGlueCatalogModel:
    def __init__(
        self,
        name: str,
        catalog_table_name: str,
        # optional properties
        external_volume: Optional[str] = None,
        catalog_namespace: Optional[str] = None,
        replace_invalid_characters: Optional[bool] = None,
        auto_refresh: Optional[bool] = None,
        catalog: Optional[str] = None,
        comment: Optional[str] = None,
    ):
        """A model object representing the IcebergTableFromAWSGlueCatalog resource.

        Constructs an object of type IcebergTableFromAWSGlueCatalog with the provided properties.

        Parameters
        __________
        name : str
            Specifies the name for the table, must be unique for the schema in which the table is created
        catalog_table_name : str
            Specifies the table name as recognized by the AWS Glue Data Catalog.
        external_volume : str, optional
            Specifies the name of the external volume to use for the table
        catalog_namespace : str, optional
            Specifies the catalog namespace for the table.
        replace_invalid_characters : bool, optional
            Specifies whether to replace invalid characters in the column names
        auto_refresh : bool, optional
            Specifies whether to automatically refresh the table metadata
        catalog : str, optional
            Specifies the name of the catalog integration to use for the table
        comment : str, optional
            Specifies a comment for the table
        """
        self.name = name
        self.external_volume = external_volume
        self.catalog_table_name = catalog_table_name
        self.catalog_namespace = catalog_namespace
        self.replace_invalid_characters = replace_invalid_characters
        self.auto_refresh = auto_refresh
        self.catalog = catalog
        self.comment = comment

    __properties = [
        "name",
        "external_volume",
        "catalog_table_name",
        "catalog_namespace",
        "replace_invalid_characters",
        "auto_refresh",
        "catalog",
        "comment",
    ]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return IcebergTableFromAWSGlueCatalog(
            name=self.name,
            external_volume=self.external_volume,
            catalog_table_name=self.catalog_table_name,
            catalog_namespace=self.catalog_namespace,
            replace_invalid_characters=self.replace_invalid_characters,
            auto_refresh=self.auto_refresh,
            catalog=self.catalog,
            comment=self.comment,
        )

    @classmethod
    def _from_model(cls, model) -> IcebergTableFromAWSGlueCatalogModel:
        return IcebergTableFromAWSGlueCatalogModel(
            name=model.name,
            external_volume=model.external_volume,
            catalog_table_name=model.catalog_table_name,
            catalog_namespace=model.catalog_namespace,
            replace_invalid_characters=model.replace_invalid_characters,
            auto_refresh=model.auto_refresh,
            catalog=model.catalog,
            comment=model.comment,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a IcebergTableFromAWSGlueCatalog.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the IcebergTableFromAWSGlueCatalog object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> IcebergTableFromAWSGlueCatalogModel:
        """Create an instance of IcebergTableFromAWSGlueCatalog from a dict.

        This method constructs a IcebergTableFromAWSGlueCatalog object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        IcebergTableFromAWSGlueCatalog
            A IcebergTableFromAWSGlueCatalog object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(IcebergTableFromAWSGlueCatalog.from_dict(obj))


IcebergTableFromAWSGlueCatalog._model_class = IcebergTableFromAWSGlueCatalogModel



from typing import Optional

import pytest

from snowflake.core import CreateMode
from snowflake.core.procedure import Procedure, ProcedureCollection, ReturnDataType, SQLFunction
from snowflake.core.tag import TagResource, TagValue

from ..base_tag_tests import BaseTagTests
from ..utils import random_string


class TestProcedureTags(BaseTagTests):
    @pytest.fixture(autouse=True)
    def setup(self, procedures: ProcedureCollection):
        self.procedures = procedures
        procedure_name = random_string(6, "procedure_for_tag_tests_")
        procedure = Procedure(
            name=procedure_name,
            arguments=[],
            return_type=ReturnDataType(datatype="VARCHAR"),
            language_config=SQLFunction(),
            body="BEGIN RETURN 'test'; END",
        )
        self.procedure_res = procedures.create(procedure, mode=CreateMode.if_not_exists)
        try:
            yield
        finally:
            self.procedure_res.drop(if_exists=True)

    @property
    def resource_level_name(self) -> str:
        return "PROCEDURE"

    def set_tags(
        self, tags: dict[TagResource, TagValue], if_exists: Optional[bool] = False, resource_name: Optional[str] = None
    ):
        self.procedures[self._normalize_resource_name(resource_name) or self.procedure_res.name_with_args].set_tags(
            tags, if_exists
        )

    def unset_tags(
        self, tag_resources: set[TagResource], if_exists: Optional[bool] = False, resource_name: Optional[str] = None
    ):
        self.procedures[self._normalize_resource_name(resource_name) or self.procedure_res.name_with_args].unset_tags(
            tag_resources, if_exists
        )

    def get_tags(
        self, with_lineage: Optional[bool] = False, resource_name: Optional[str] = None
    ) -> dict[TagResource, TagValue]:
        return self.procedures[
            self._normalize_resource_name(resource_name) or self.procedure_res.name_with_args
        ].get_tags(with_lineage)

    @staticmethod
    def _normalize_resource_name(name: str | None) -> str | None:
        return f"{name}()" if name and not name.endswith(")") else name



import pytest

from snowflake.core.database_role import ContainingScope, Securable
from snowflake.core.schema import Schema
from snowflake.core.table import Table, TableColumn
from tests.utils import random_string


pytestmark = pytest.mark.min_sf_ver("8.39.0")


@pytest.mark.use_accountadmin
def test_grant_role(database_roles_in_test_database, test_database_role_name, test_database_role_name_2):
    database_roles_in_test_database[test_database_role_name].grant_role(
        role_type="DATABASE ROLE", role=Securable(name=test_database_role_name_2)
    )

    # running the same grant again should not be an issue
    database_roles_in_test_database[test_database_role_name].grant_role(
        role_type="DATABASE ROLE", role=Securable(name=test_database_role_name_2)
    )

    grants = list(database_roles_in_test_database[test_database_role_name].iter_grants_to())
    assert len(grants) == 2  # a grant of usage on database is by default granted to a database role
    for grant in grants:
        assert grant.securable_type in ["DATABASE", "DATABASE ROLE"]


@pytest.mark.use_accountadmin
def test_grant_privileges(
    database_roles_in_test_database,
    test_database_role_name,
    test_database_for_grant_name,
    test_schema_for_grant_name,
    test_table_for_grant_name,
):
    # grant on a specific database
    database_roles_in_test_database[test_database_role_name].grant_privileges(
        privileges=["MODIFY", "MONITOR"],
        securable_type="DATABASE",
        securable=Securable(name=test_database_for_grant_name),
    )

    # grant on a schema
    database_roles_in_test_database[test_database_role_name].grant_privileges(
        privileges=["CREATE TASK"],
        securable_type="SCHEMA",
        securable=Securable(database=test_database_for_grant_name, name=test_schema_for_grant_name),
    )

    # grant on a table
    database_roles_in_test_database[test_database_role_name].grant_privileges(
        privileges=["SELECT"],
        securable_type="TABLE",
        securable=Securable(
            database=test_database_for_grant_name, schema=test_schema_for_grant_name, name=test_table_for_grant_name
        ),
    )

    assert len(list(database_roles_in_test_database[test_database_role_name].iter_grants_to())) == 5


@pytest.mark.use_accountadmin
def test_grant_privileges_with_grant_option(
    database_roles_in_test_database,
    test_database_role_name,
    test_database_for_grant_name,
    test_schema_for_grant_name,
    test_table_for_grant_name,
):
    # grant on a specific database
    database_roles_in_test_database[test_database_role_name].grant_privileges(
        privileges=["MODIFY", "MONITOR"],
        securable_type="DATABASE",
        securable=Securable(name=test_database_for_grant_name),
        grant_option=True,
    )

    # grant on a schema
    database_roles_in_test_database[test_database_role_name].grant_privileges(
        privileges=["CREATE TASK"],
        securable_type="SCHEMA",
        securable=Securable(database=test_database_for_grant_name, name=test_schema_for_grant_name),
        grant_option=True,
    )

    # grant on a table
    database_roles_in_test_database[test_database_role_name].grant_privileges(
        privileges=["SELECT"],
        securable_type="TABLE",
        securable=Securable(
            database=test_database_for_grant_name, schema=test_schema_for_grant_name, name=test_table_for_grant_name
        ),
        grant_option=True,
    )

    grant_list = list(database_roles_in_test_database[test_database_role_name].iter_grants_to())
    assert len(grant_list) == 5

    for grant in grant_list:
        if not (grant.securable_type == "DATABASE" and grant.privileges == ["USAGE"]):
            assert grant.grant_option


@pytest.mark.use_accountadmin
def test_grant_privileges_on_all(
    test_database_role_name,
    test_database_for_grant_name,
    test_schema_for_grant_name,
    test_table_for_grant_name,
    database_roles_in_test_database,
):
    del test_table_for_grant_name
    # all schemas in a database
    database_roles_in_test_database[test_database_role_name].grant_privileges_on_all(
        privileges=["USAGE"],
        securable_type="SCHEMA",
        containing_scope=ContainingScope(database=test_database_for_grant_name),
    )

    # all tables in a schema
    database_roles_in_test_database[test_database_role_name].grant_privileges_on_all(
        privileges=["SELECT"],
        securable_type="TABLE",
        containing_scope=ContainingScope(database=test_database_for_grant_name, schema=test_schema_for_grant_name),
    )

    grants = list(database_roles_in_test_database[test_database_role_name].iter_grants_to())
    for grant in grants:
        if not (grant.securable_type == "DATABASE" and grant.privileges == ["USAGE"]):
            if grant.securable_type == "SCHEMA":
                assert grant.securable.database == test_database_for_grant_name.upper()
                assert grant.securable.name is not None
            else:
                assert grant.securable_type == "TABLE"
                assert grant.securable.database == test_database_for_grant_name.upper()
                assert grant.securable.var_schema == test_schema_for_grant_name.upper()
                assert grant.securable.name is not None
    assert len(grants) >= 3


@pytest.mark.use_accountadmin
def test_grant_future_privileges(
    database_roles_in_test_database,
    test_database_role_name,
    test_database_for_grant_name,
    test_schema_for_grant_name,
    backup_database_schema,
    schemas_in_test_database,
    tables_in_test_database,
):
    del backup_database_schema
    # all future schemas in a database
    database_roles_in_test_database[test_database_role_name].grant_future_privileges(
        privileges=["USAGE"],
        securable_type="SCHEMA",
        containing_scope=ContainingScope(database=test_database_for_grant_name),
    )

    # all future tables in a schema
    database_roles_in_test_database[test_database_role_name].grant_future_privileges(
        privileges=["INSERT"],
        securable_type="TABLE",
        containing_scope=ContainingScope(database=test_database_for_grant_name, schema=test_schema_for_grant_name),
    )

    assert len(list(database_roles_in_test_database[test_database_role_name].iter_grants_to())) == 1

    # create a new schema and table
    try:
        schema_name = random_string(4, "test_grant_future_privileges_schema_")
        schemas_in_test_database.create(Schema(name=schema_name))
        table_name = random_string(5, "test_grant_future_privileges_table_")
        columns = [TableColumn(name="col1", datatype="int"), TableColumn(name="col2", datatype="string")]
        test_table = Table(name=table_name, columns=columns)
        tables_in_test_database.create(test_table)

        assert len(list(database_roles_in_test_database[test_database_role_name].iter_grants_to())) == 3
    finally:
        schemas_in_test_database[schema_name].drop()
        tables_in_test_database[table_name].drop()



# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.

import os
import shutil
import tempfile
import typing

from contextlib import contextmanager

import snowflake.connector

from ..utils import is_prod_or_preprod, random_string


if typing.TYPE_CHECKING:
    from snowflake.snowpark import Row, Session


MASKED_VALUE = "********"


def random_object_name() -> str:
    return random_string(8, prefix="test_object_")


def get_task_history(session: "Session", name: str) -> list["Row"]:
    query = (
        f"select * from table(information_schema.task_history("
        f"scheduled_time_range_start=>dateadd('hour',-1,current_timestamp()),"
        f"result_limit => 10,task_name=>'{name}'))"
    )
    return session.sql(query).collect()


def string_skip_space_and_cases(s):
    return s.replace(" ", "").upper()


def array_equal_comparison(arr1, arr2):
    if not arr1 and not arr2:
        return True
    if not arr1 or not arr2:
        return False

    return [string_skip_space_and_cases(i) for i in arr1] == [string_skip_space_and_cases(i) for i in arr2]


def connection_config(override_schema=None, override_database=None, connection_name=None, error_if_not_exists=True):
    config = {}
    try:
        from ..parameters import CONNECTION_PARAMETERS
    except ImportError:
        CONNECTION_PARAMETERS = None
        from snowflake.connector.config_manager import CONFIG_MANAGER

    if CONNECTION_PARAMETERS is None:
        if connection_name is None:
            connection_key = CONFIG_MANAGER["default_connection_name"]
        else:
            connection_key = connection_name

        # 2023-06-23(warsaw): By default, we read out of the [connections.snowflake] section in the config.toml file,
        # but by setting the environment variable SNOWFLAKE_DEFAULT_CONNECTION_NAME you can read out of a different
        # section. For example SNOWFLAKE_DEFAULT_CONNECTION_NAME='test' reads out of [connections.test]
        if connection_key not in CONFIG_MANAGER["connections"]:
            if error_if_not_exists:
                raise KeyError("Connection config is missing.")
            else:
                return None

        config = CONFIG_MANAGER["connections"][connection_key]
    else:
        config = CONNECTION_PARAMETERS

    if override_schema:
        config["schema"] = override_schema
    if override_database:
        config["database"] = override_database
    return config


def should_disable_setup_for_spcs(config):
    return config.get("should_disable_setup_for_spcs", "") == "true"


def should_disable_setup_for_credentials(config):
    return config.get("should_disable_setup_for_credentials", "") == "true"


def get_snowflake_version(cursor):
    return cursor.execute("SELECT CURRENT_VERSION()").fetchone()[0].strip()


def get_snowflake_region(cursor):
    return cursor.execute("SELECT CURRENT_REGION()").fetchone()[0].strip()


def connection_keys():
    return [
        "user",
        "password",
        "host",
        "port",
        "database",
        "schema",
        "account",
        "protocol",
        "role",
        "warehouse",
        "private_key_file",
        "private_key_file_pwd",
    ]


@contextmanager
def backup_role(cursor):
    _current_role = cursor.execute("SELECT /* use_role */ CURRENT_ROLE()").fetchone()[0]
    try:
        yield
    finally:
        if _current_role is not None:
            cursor.execute(f"USE ROLE /* use_role::reset */ {_current_role}").fetchone()


@contextmanager
def backup_database_and_schema(cursor):
    _current_database = cursor.execute("SELECT CURRENT_DATABASE()").fetchone()[0]
    _current_schema = cursor.execute("SELECT CURRENT_SCHEMA()").fetchone()[0]
    try:
        yield
    finally:
        if _current_database is not None:
            cursor.execute(f"USE DATABASE /* use_database::reset */ {_current_database}").fetchone()
        if _current_schema is not None:
            cursor.execute(f"USE SCHEMA /* use_schema::reset */ {_current_schema}").fetchone()


@contextmanager
def backup_warehouse(cursor):
    _current_warehouse = cursor.execute("SELECT CURRENT_WAREHOUSE()").fetchone()[0]
    try:
        yield _current_warehouse
    finally:
        if _current_warehouse is not None:
            cursor.execute(f"USE WAREHOUSE /* use_warehouse::reset */ {_current_warehouse}").fetchone()


def create_zip_from_paths(paths, output_filename):
    try:
        with tempfile.TemporaryDirectory() as temp_dir:
            for path in paths:
                if os.path.isdir(path):
                    folder_name = os.path.basename(path)
                    temp_folder = os.path.join(temp_dir, folder_name)
                    shutil.copytree(path, temp_folder)
                elif os.path.isfile(path):
                    shutil.copy(path, temp_dir)
                else:
                    print(f"Warning: '{path}' is not a valid file or directory. Skipping.")

            shutil.make_archive(os.path.splitext(output_filename)[0], "zip", root_dir=temp_dir)
    except Exception as e:
        raise Exception(f"Error creating the snowflake core zip file:\n {e.with_traceback(None)}") from e


def create_and_use_new_database_and_schema(cursor, new_database_name, new_schema_name):
    # Database
    cursor.execute(f"CREATE DATABASE IF NOT EXISTS /* setup_basic */ {new_database_name} DATA_RETENTION_TIME_IN_DAYS=1")
    cursor.execute(f"USE DATABASE {new_database_name}")

    # Schema
    cursor.execute(f"CREATE SCHEMA IF NOT EXISTS {new_schema_name}")
    cursor.execute(f"USE SCHEMA {new_schema_name}")


def upload_given_files_to_stage(cursor, stage_url, files):
    try:
        for file in files:
            cursor.execute(f"PUT file://{file} @{stage_url} AUTO_COMPRESS=FALSE OVERWRITE=TRUE")
    except Exception as e:
        raise Exception(f"Error uploading files to the stage:\n {e.with_traceback(None)}") from e


def execute_notebook(cursor, notebook_name, stage_full_path, warehouse_name, notebook_file_name) -> bool:
    try:
        cursor.execute(
            f"CREATE OR REPLACE NOTEBOOK {notebook_name} "
            f"FROM '@{stage_full_path}' "
            f"MAIN_FILE = '{notebook_file_name}' QUERY_WAREHOUSE = {warehouse_name}"
        )
        cursor.execute(f"ALTER NOTEBOOK {notebook_name} ADD LIVE VERSION FROM LAST")
        cursor.execute(f"EXECUTE NOTEBOOK {notebook_name}()")
        return False
    except Exception as e:
        print(f"Error creating and executing the notebook file {notebook_file_name}:\n {e.with_traceback(None)}")
        return True


def setup_spcs(target_account_name=None, executing_account_cursor=None, instance_families_to_create=None):
    if instance_families_to_create is None:
        instance_families_to_create = ["FAKE"]
    if target_account_name is not None:
        prefix = f"alter account {target_account_name} "
    else:
        prefix = "alter account "
    executing_account_cursor.execute(f"{prefix} set snowservices_external_image_registry_allowlist = '*';").fetchone()
    executing_account_cursor.execute(f"{prefix} set enable_snowservices=true;").fetchone()
    executing_account_cursor.execute(f"{prefix} set enable_snowservices_user_facing_features=true;").fetchone()


def setup_credentials(cursor):
    add_cred_to_pool_cmd = {
        "AWS": f"""
            select system$creds_add_credentials_to_pool('COMMON_AWS_INTEGRATION',
            'AWS',
            '[{{"AWS_KEY_ID": "{os.environ["AWS_ACCESS_KEY_ID"]}",
                "AWS_SECRET_KEY": "{os.environ["AWS_SECRET_ACCESS_KEY"]}",
                "CREDENTIAL_NAME": "aws-creds",
                "CREDENTIAL_TYPE": "AWS_IAM_USER"}}]',
            '{{"PROVIDER_SPECIFIC_NAME":"AWS-specific data"}}')
        """,
        "AZURE": f"""
            select system$creds_add_credentials_to_pool('COMMON_AZURE_INTEGRATION',
            'AZURE',
            '[{{"AZURE_CLIENT_ID": "{os.environ["AZURE_AD_APP_CLIENT_ID"]}",
                "AZURE_CLIENT_SECRET": "{os.environ["AZURE_AD_APP_CLIENT_SECRET"]}",
                "CREDENTIAL_NAME": "azure-creds",
                "AZURE_TENANT_ID": "{os.environ["AZURE_AD_APP_TENANT_ID"]}",
                "CREDENTIAL_TYPE": "AZURE_APP"}}]',
            '{{"PROVIDER_SPECIFIC_NAME":"AZURE-specific data"}}')
        """,
        "GOOGLE": f"""
            select system$creds_add_credentials_to_pool('COMMON_GOOGLE_INTEGRATION',
            'GOOGLE',
            '[{{"GCS_SERVICE_ACCOUNT_KEY_BASE64": "{os.environ["SITE_GCS_SERVICE_ACCOUNT_KEY_BASE64"]}",
                "CREDENTIAL_NAME": "gcp-creds",
                "CREDENTIAL_TYPE": "GCP_SERVICE_ACCOUNT"}}]',
            '{{"PROVIDER_SPECIFIC_NAME":"GOOGLE-specific data"}}')
        """,
    }
    cursor.execute(add_cred_to_pool_cmd["AWS"])
    cursor.execute(add_cred_to_pool_cmd["AZURE"])
    cursor.execute(add_cred_to_pool_cmd["GOOGLE"])


def setup_account_for_notebook(cursor, config):
    # if it's a prod or a preprod account there shouldn't be requirement to do this setup
    if is_prod_or_preprod(get_snowflake_version(cursor), get_snowflake_region(cursor)):
        return

    sf_connection_parameters = connection_config(connection_name="sf_account", error_if_not_exists=False)

    if config["account"] != "snowflake" and sf_connection_parameters is None:
        raise Exception("Account is not snowflake or prod and sf_account connection parameters are not provided")

    if config["account"] == "snowflake":
        setup_spcs(executing_account_cursor=cursor, instance_families_to_create=["CPU_X64_XS", "FAKE"])
        cursor.execute("ALTER ACCOUNT SET FEATURE_NOTEBOOKS_NON_INTERACTIVE_EXECUTION = 'ENABLED';")
        setup_credentials(cursor)
    else:
        _keys = connection_keys()
        with snowflake.connector.connect(
            **{k: sf_connection_parameters[k] for k in _keys if k in sf_connection_parameters}
        ) as sf_conn:
            target_account_name = config["account"]
            setup_spcs(
                executing_account_cursor=sf_conn.cursor(),
                target_account_name=target_account_name,
                instance_families_to_create=["CPU_X64_XS", "FAKE"],
            )
            sf_conn.cursor().execute(
                f"ALTER ACCOUNT {target_account_name} SET FEATURE_NOTEBOOKS_NON_INTERACTIVE_EXECUTION = 'ENABLED';"
            )
            setup_credentials(sf_conn.cursor())



from concurrent.futures import Future
from typing import TYPE_CHECKING, Literal, Optional, Union, overload

from pydantic import StrictStr

from snowflake.core import FQN, PollingOperation
from snowflake.core._common import (
    Clone,
    CreateMode,
)

from .._internal.telemetry import api_telemetry
from ._generated import SuccessResponse
from ._generated.api.stream_api_base import StreamCollectionBase, StreamResourceBase
from ._generated.models.stream import Stream
from ._generated.models.stream_clone import StreamClone


if TYPE_CHECKING:
    from snowflake.core.schema import SchemaResource


class StreamCollection(StreamCollectionBase):
    """Represents the collection operations on the Snowflake Stream resource.

    With this collection, you can create, iterate through, and fetch streams
    that you have access to in the current context.

    Examples
    ________
    Creating a stream instance:

    >>> streams = root.databases["my_db"].schemas["my_schema"].streams
    >>> streams.create(
    ...     Stream(
    ...         name="my_stream",
    ...         stream_source=StreamSourceTable(name="my_table", append_only=True, show_initial_rows=False),
    ...     ),
    ...     mode=CreateMode.error_if_exists,
    ... )
    """

    def __init__(self, schema: "SchemaResource"):
        super().__init__(schema, StreamResource)

    @overload  # type: ignore[override]
    @api_telemetry
    def create(
        self,
        stream: str,
        *,
        clone_stream: Union[str, Clone],
        mode: CreateMode = CreateMode.error_if_exists,
        copy_grants: Optional[bool] = False,
    ) -> "StreamResource": ...
    @overload
    @api_telemetry
    def create(
        self,
        stream: Stream,
        *,
        mode: CreateMode = CreateMode.error_if_exists,
        copy_grants: Optional[bool] = False,
    ) -> "StreamResource": ...
    @api_telemetry
    def create(
        self,
        stream: Union[str, Stream],
        *,
        clone_stream: Optional[Union[str, Clone]] = None,
        mode: CreateMode = CreateMode.error_if_exists,
        copy_grants: Optional[bool] = False,
    ) -> "StreamResource":
        """Create a stream in Snowflake.

        There are two ways to create a stream: by cloning or by building from scratch.

        **Cloning an existing stream**

        Parameters
        __________
        stream: str
            The new stream's name
        clone_stream: str or Clone object
            The name of stream to be cloned
        mode: CreateMode, optional
            One of the following enum values:

            ``CreateMode.error_if_exists``: Throw an :class:`snowflake.core.exceptions.ConflictError`
            if the stream already exists in Snowflake.  Equivalent to SQL ``create stream <name> ...``.

            ``CreateMode.or_replace``: Replace if the stream already exists in Snowflake. Equivalent to SQL
            ``create or replace stream <name> ...``.

            ``CreateMode.if_not_exists``: Do nothing if the stream already exists in Snowflake.
            Equivalent to SQL ``create stream <name> if not exists...``

            Default is ``CreateMode.error_if_exists``.

        copy_grants: bool, optional
            Whether to enable copy grants when creating the object. Default is ``False``.

        Examples
        ________
        Cloning a Stream instance:

        >>> streams = schema.streams
        >>> streams.create(
        ...     "new_stream_name",
        ...     clone_stream="stream_name_to_be_cloned",
        ...     mode=CreateMode.if_not_exists,
        ...     copy_grants=True,
        ... )

        Cloning a Stream instance in a different database and schema

        >>> streams = schema.streams
        >>> streams.create(
        ...     "new_stream_name",
        ...     clone_stream="stream_database_name.stream_schema_name.stream_name_to_be_cloned",
        ...     mode=CreateMode.if_not_exists,
        ...     copy_grants=True,
        ... )

        **Creating a stream from scratch**

        Parameters
        __________
        stream: Stream
            The details of ``Stream`` object, together with ``Stream``'s properties:
            name; comment is optional
            stream_source: ``StreamSource`` object, one of:
            ``StreamSourceStage``, ``StreamSourceTable``, ``StreamSourceView``.

        mode: CreateMode, optional
            One of the following enum values:

            ``CreateMode.error_if_exists``: Throw an :class:`snowflake.core.exceptions.ConflictError`
            if the stream already exists in Snowflake.  Equivalent to SQL ``create stream <name> ...``.

            ``CreateMode.or_replace``: Replace if the stream already exists in Snowflake. Equivalent to SQL
            ``create or replace stream <name> ...``.

            ``CreateMode.if_not_exists``: Do nothing if the stream already exists in Snowflake.
            Equivalent to SQL ``create stream <name> if not exists...``

            Default is ``CreateMode.error_if_exists``.

        copy_grants: bool, optional
            Whether to enable copy grants when creating the object. Default is ``False``.

        Examples
        ________
        Creating a stream instance by source table:

        >>> streams.create(
        ...     Stream(
        ...         name = "new_stream_name",
        ...         stream_source = StreamSourceTable(
        ...             point_of_time = PointOfTimeOffset(reference="before", offset="1"),
        ...             name = "my_source_table_name"
        ...             append_only = True,
        ...             show_initial_rows = False,
        ...             comment = "create stream by table"
        ...         )
        ...     ),
        ...     mode=CreateMode.if_not_exists,
        ...     copy_grants=True
        ... )

        Creating a stream instance by source view:

        >>> streams.create(
        ...     Stream(
        ...         name="new_stream_name",
        ...         stream_source=StreamSourceView(
        ...             point_of_time=PointOfTimeOffset(reference="before", offset="1"),
        ...             name="my_source_view_name",
        ...         ),
        ...     ),
        ...     mode=CreateMode.if_not_exists,
        ...     copy_grants=True,
        ... )

        Creating a stream instance by source directory table:

        >>> streams.create(
        ...     Stream(
        ...         name="new_stream_name",
        ...         stream_source=StreamSourceStage(
        ...             point_of_time=PointOfTimeOffset(reference="before", offset="1"),
        ...             name="my_source_directory_table_name",
        ...         ),
        ...     ),
        ...     mode=CreateMode.if_not_exists,
        ...     copy_grants=True,
        ... )
        """
        self._create(
            stream=stream,
            clone_stream=clone_stream,
            mode=mode,
            copy_grants=copy_grants,
            async_req=False,
        )
        return StreamResource(stream.name if isinstance(stream, Stream) else stream, self)

    @overload  # type: ignore[override]
    @api_telemetry
    def create_async(
        self,
        stream: str,
        *,
        clone_stream: Union[str, Clone],
        mode: CreateMode = CreateMode.error_if_exists,
        copy_grants: Optional[bool] = False,
    ) -> PollingOperation["StreamResource"]: ...

    @overload
    @api_telemetry
    def create_async(
        self,
        stream: Stream,
        *,
        mode: CreateMode = CreateMode.error_if_exists,
        copy_grants: Optional[bool] = False,
    ) -> PollingOperation["StreamResource"]: ...

    @api_telemetry
    def create_async(
        self,
        stream: Union[str, Stream],
        *,
        clone_stream: Optional[Union[str, Clone]] = None,
        mode: CreateMode = CreateMode.error_if_exists,
        copy_grants: Optional[bool] = False,
    ) -> PollingOperation["StreamResource"]:
        """An asynchronous version of :func:`create`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self._create(
            stream=stream,
            clone_stream=clone_stream,
            mode=mode,
            copy_grants=copy_grants,
            async_req=True,
        )
        return PollingOperation(
            future, lambda _: StreamResource(stream.name if isinstance(stream, Stream) else stream, self)
        )

    @overload
    def _create(
        self,
        stream: Union[str, Stream],
        clone_stream: Optional[Union[str, Clone]],
        mode: CreateMode,
        copy_grants: Optional[bool],
        async_req: Literal[True],
    ) -> Future[SuccessResponse]: ...

    @overload
    def _create(
        self,
        stream: Union[str, Stream],
        clone_stream: Optional[Union[str, Clone]],
        mode: CreateMode,
        copy_grants: Optional[bool],
        async_req: Literal[False],
    ) -> SuccessResponse: ...

    def _create(
        self,
        stream: Union[str, Stream],
        clone_stream: Optional[Union[str, Clone]],
        mode: CreateMode,
        copy_grants: Optional[bool],
        async_req: bool,
    ) -> Union[SuccessResponse, Future[SuccessResponse]]:
        real_mode = CreateMode[mode].value

        if clone_stream:
            # create stream by clone
            if not isinstance(stream, str):
                raise TypeError("Stream has to be str for clone")

            real_clone = Clone(source=clone_stream) if isinstance(clone_stream, str) else clone_stream

            req = StreamClone(
                name=stream,
            )

            source_stream_fqn = FQN.from_string(real_clone.source)
            return self._api.clone_stream(
                source_stream_fqn.database or self.database.name,
                source_stream_fqn.schema or self.schema.name,
                source_stream_fqn.name,
                create_mode=StrictStr(real_mode),
                target_database=self.database.name,
                target_schema=self.schema.name,
                stream_clone=req,
                copy_grants=copy_grants,
                async_req=async_req,
            )

        if not isinstance(stream, Stream):
            raise TypeError("Stream has to be Stream object")
        return self._api.create_stream(
            self.database.name,
            self.schema.name,
            stream,
            create_mode=StrictStr(real_mode),
            copy_grants=copy_grants,
            async_req=async_req,
        )


class StreamResource(StreamResourceBase):
    """Represents a reference to a Snowflake Stream resource.

    With this stream reference, you can fetch information about a stream, as well as perform
    certain actions on it.
    """

    _plural_name = "streams"



"""Helpers related to sending/receiving HTTP requests."""

import datetime
import json
import logging
import os
import ssl
import typing

from decimal import Decimal
from typing import Optional
from urllib.parse import quote

import urllib3

from pydantic import SecretStr

from snowflake.core.exceptions import SnowflakePythonError

from ._common import TokenType
from ._constants import SESSION_TOKEN_EXPIRED_ERROR_CODE


if typing.TYPE_CHECKING:
    import snowflake.core

    from ._generated import Configuration

logger = logging.getLogger(__name__)

PrimitiveTypes = typing.TypeVar("PrimitiveTypes", float, bool, bytes, bytearray, str, int)
PRIMITIVE_TYPES = (float, bool, bytes, bytearray, str, int)
NATIVE_TYPES_MAPPING = {
    "int": int,
    "long": int,  # TODO remove as only py3 is supported?
    "float": float,
    "str": str,
    "bool": bool,
    "date": datetime.date,
    "datetime": datetime.datetime,
    "object": object,
    "bytes": bytes,
    "bytearray": bytearray,
}
DEFAULT_RETRY_TIMEOUT_SECONDS = 600.0  # default 10 minutes for query retries
STATUS_CODES_MAPPING = {
    200: "OK",
    202: "Long Running Query",
    429: "Rate Limited",
    503: "Service Unavailable",
    504: "Gateway Timeout",
}


def resolve_url(
    resource_path: str, path_params: dict[str, str], collection_formats: dict[typing.Any, typing.Any], safe_quoting: str
) -> str:
    if path_params:
        path_params = sanitize_for_serialization(path_params)
        path_params_list = parameters_to_tuples(path_params, collection_formats)
        resource_path = resource_path.format(**{k: quote(str(v), safe=safe_quoting) for k, v in path_params_list})
    return resource_path


class ToDictProto(typing.Protocol):
    def to_dict(self) -> dict[typing.Any, typing.Any]: ...


@typing.overload
def sanitize_for_serialization(obj: None) -> None: ...
@typing.overload
def sanitize_for_serialization(obj: PrimitiveTypes) -> PrimitiveTypes: ...
@typing.overload
def sanitize_for_serialization(obj: Decimal) -> str: ...
@typing.overload
def sanitize_for_serialization(obj: list[typing.Any]) -> list[typing.Any]: ...
@typing.overload
def sanitize_for_serialization(obj: tuple[typing.Any, ...]) -> tuple[typing.Any, ...]: ...
@typing.overload
def sanitize_for_serialization(obj: typing.Union[datetime.datetime, datetime.date]) -> str: ...
@typing.overload
def sanitize_for_serialization(obj: dict[typing.Any, typing.Any]) -> dict[typing.Any, typing.Any]: ...
@typing.overload
def sanitize_for_serialization(obj: ToDictProto) -> dict[typing.Any, typing.Any]: ...
def sanitize_for_serialization(
    obj: typing.Union[
        None,
        float,
        bool,
        bytes,
        str,
        int,
        Decimal,
        list[typing.Any],
        tuple[typing.Any, ...],
        datetime.datetime,
        datetime.date,
        dict[typing.Any, typing.Any],
        ToDictProto,
    ],
) -> typing.Union[
    None, float, bool, bytes, str, int, list[typing.Any], tuple[typing.Any, ...], dict[typing.Any, typing.Any]
]:
    """Build a JSON POST object.

    If obj is None, return None.
    If obj is str, int, long, float, bool, return directly.
    if obj is decimal.Decimal, convert to string, using scientific notation if needed.
    If obj is datetime.datetime, datetime.date
        convert to string in iso8601 format.
    If obj is list, sanitize each element in the list.
    If obj is dict, return the dict.
    If obj is OpenAPI model, return the properties dict.

    :param obj: The data to serialize.
    :return: The serialized form of data.
    """
    if obj is None:
        return None
    elif isinstance(obj, PRIMITIVE_TYPES):
        return obj
    elif isinstance(obj, Decimal):
        return str(obj)
    elif isinstance(obj, SecretStr):
        return obj.get_secret_value()
    elif isinstance(obj, list):
        return [sanitize_for_serialization(sub_obj) for sub_obj in obj]
    elif isinstance(obj, tuple):
        return tuple(sanitize_for_serialization(sub_obj) for sub_obj in obj)
    elif isinstance(obj, (datetime.datetime, datetime.date)):
        return obj.isoformat()

    if isinstance(obj, dict):
        obj_dict = obj
    else:
        # Convert model obj to dict except
        # attributes `openapi_types`, `attribute_map`
        # and attributes which value is not None.
        # Convert attribute name to json key in
        # model definition for request.
        # Prefer to use `to_dict_without_readonly_properties` if the model has
        if hasattr(obj, "to_dict_without_readonly_properties"):
            obj_dict = obj.to_dict_without_readonly_properties()
        else:
            obj_dict = obj.to_dict()

    return {key: sanitize_for_serialization(val) for key, val in obj_dict.items()}


def parameters_to_tuples(
    params: typing.Union[dict[typing.Any, typing.Any], list[tuple[typing.Any, typing.Any]]],
    collection_formats: typing.Optional[dict[typing.Any, typing.Any]],
) -> list[tuple[typing.Any, typing.Any]]:
    """Get parameters as list of tuples, formatting collections.

    :param params: Parameters as dict or list of two-tuples
    :param dict collection_formats: Parameter collection formats
    :return: Parameters as list of tuples, collections formatted
    """
    new_params: list[tuple[typing.Any, typing.Any]] = []
    if collection_formats is None:
        collection_formats = {}
    for k, v in params.items() if isinstance(params, dict) else params:
        if k in collection_formats:
            collection_format = collection_formats[k]
            if collection_format == "multi":
                new_params.extend((k, value) for value in v)
            else:
                if collection_format == "ssv":
                    delimiter = " "
                elif collection_format == "tsv":
                    delimiter = "\t"
                elif collection_format == "pipes":
                    delimiter = "|"
                else:  # csv is the default
                    delimiter = ","
                new_params.append((k, delimiter.join(str(value) for value in v)))
        else:
            new_params.append((k, v))
    return new_params


class SFPoolManager:
    def __init__(  # type: ignore[no-untyped-def]
        self,
        manager_class: typing.Union[type[urllib3.ProxyManager], type[urllib3.PoolManager]],
        **cp_kwargs,
    ):
        self._manager = manager_class(**cp_kwargs)

    # Having this typed is non-trivial across multiple
    #  urllib3 major versions
    def request(  # type: ignore[no-untyped-def]
        self,
        root: "snowflake.core.Root",
        method: str,
        url: str,
        fields=None,
        headers: typing.Optional[dict[str, str]] = None,
        **urlopen_kw: typing.Any,
    ):
        if headers is None:
            headers = dict()
        need_auth = url_needs_auth(url)
        if need_auth:
            if root._session_token is None:
                # This should never trigger
                raise Exception("session token is missing while making a request")
            headers.update(get_session_headers(root.token_type, root._session_token, root.external_session_id))
        logger.debug("making an http %s call to '%s'", method.upper(), url)
        try:
            r = self._manager.request(method=method, url=url, fields=fields, headers=headers, **urlopen_kw)
        except urllib3.exceptions.MaxRetryError as e:
            if (
                isinstance(e.reason, urllib3.exceptions.SSLError)
                and isinstance(e.reason.args[0], ssl.SSLCertVerificationError)
                and "Hostname mismatch" in e.reason.args[0].verify_message
            ):
                raise SnowflakePythonError(
                    "This SSL error occurs when the hostname contains underscores, please see "
                    "https://docs.snowflake.com/en/user-guide/organizations-connect for how to set "
                    "the hostname correctly."
                ) from e
            raise

        try:
            content_type = r.headers.get("Content-Type")
            if content_type in {"text/event-stream", "application/octet-stream"}:
                resp_json = dict()
            else:
                resp_json = json.loads(r.data)
        except Exception:
            resp_json = dict()
        if (
            need_auth
            and isinstance(resp_json, dict)
            and resp_json.get("error_code") == SESSION_TOKEN_EXPIRED_ERROR_CODE
            and hasattr(root, "_connection")
            and root._connection.rest is not None
        ):
            # Try renewing session token and try request again
            logger.debug("session expired, renewing session")
            root._connection.rest._renew_session()
            if need_auth:
                if root._session_token is None:
                    # This should never trigger
                    raise Exception("session token is missing right after renewal")
                headers.update(get_session_headers(root.token_type, root._session_token, root.external_session_id))
            logger.debug("repeating an http with new session token %s call to '%s'", method.upper(), url)
            r = self._manager.request(method=method, url=url, fields=fields, headers=headers, **urlopen_kw)
        return r


# Use a connection pool singleton for every resource
CONNECTION_POOL: typing.Optional[SFPoolManager] = None


def get_session_headers(
    token_type: TokenType, session_token: str, external_session_id: Optional[str] = None
) -> dict[str, str]:
    if token_type is TokenType.EXTERNAL_SESSION_WITH_PAT:
        return {
            "Authorization": f"Bearer {session_token}",
            "X-Snowflake-External-Session-ID": external_session_id or "",
            "X-Snowflake-Authorization-Token-Type": "PAT_WITH_EXTERNAL_SESSION_ID",
        }
    return {"Authorization": f'Snowflake Token="{session_token}"'}


def url_needs_auth(url: str) -> bool:
    """Whether a URL needs the authentication headers to work.

    For now this is all URLs, since Python connector takes care of authentication and
    session related actions.
    """
    return True


def _proxy_setup(configuration: "Configuration") -> Optional[dict[str, typing.Any]]:
    if configuration.proxy:
        return {"proxy_url": configuration.proxy, "proxy_headers": configuration.proxy_headers}
    # To keep compatibility with driver behavior
    if http_url := os.getenv("HTTPS_PROXY"):
        return {"proxy_url": http_url, "proxy_headers": {}}
    return None


# TODO: We could create the single connection pool at import time
#  instead of having this function at all
# TODO: Configuration classes have no single parent class
def create_connection_pool(  # type: ignore[no-untyped-def]
    configuration, pools_size: int = 4, maxsize: typing.Optional[int] = None
) -> SFPoolManager:
    # TODO: locking?
    global CONNECTION_POOL
    if CONNECTION_POOL is None:
        # urllib3.PoolManager will pass all kw parameters to connectionpool
        # https://github.com/shazow/urllib3/blob/f9409436f83aeb79fbaf090181cd81b784f1b8ce/urllib3/poolmanager.py#L75  # noqa: E501
        # https://github.com/shazow/urllib3/blob/f9409436f83aeb79fbaf090181cd81b784f1b8ce/urllib3/connectionpool.py#L680  # noqa: E501
        # maxsize is the number of requests to host that are allowed in parallel
        # Custom SSL certificates and client certificates: http://urllib3.readthedocs.io/en/latest/advanced-usage.html  # noqa: E501

        # cert_reqs
        if configuration.verify_ssl:
            cert_reqs = ssl.CERT_REQUIRED
        else:
            cert_reqs = ssl.CERT_NONE

        addition_pool_args = {}
        if configuration.assert_hostname is not None:
            addition_pool_args["assert_hostname"] = configuration.assert_hostname

        if configuration.retries is not None:
            addition_pool_args["retries"] = configuration.retries

        if configuration.socket_options is not None:
            addition_pool_args["socket_options"] = configuration.socket_options

        if maxsize is None:
            if configuration.connection_pool_maxsize is not None:
                maxsize = configuration.connection_pool_maxsize
            else:
                maxsize = 4

        # https pool manager
        cp_kwargs = {
            "num_pools": pools_size,
            "maxsize": maxsize,
            "cert_reqs": cert_reqs,
            "ca_certs": configuration.ssl_ca_cert,
            "cert_file": configuration.cert_file,
            "key_file": configuration.key_file,
            **addition_pool_args,
        }
        if proxy_kw := _proxy_setup(configuration):
            manager_class = urllib3.ProxyManager
            cp_kwargs.update(proxy_kw)
        else:
            manager_class = urllib3.PoolManager  # type: ignore[assignment]
        logger.debug("created a new SFPoolManager")
        CONNECTION_POOL = SFPoolManager(manager_class=manager_class, **cp_kwargs)
    return CONNECTION_POOL



from collections.abc import Iterator

import pytest

from snowflake.core.api_integration import ApiIntegration, GitHook
from tests.integ.utils import random_string


@pytest.fixture(scope="module")
def secret(cursor) -> Iterator[str]:
    secret_name = random_string(10, "test_secret_")
    cursor.execute(f"CREATE SECRET IDENTIFIER('{secret_name}') TYPE = PASSWORD USERNAME = 'admin' PASSWORD = 'test'")
    try:
        yield secret_name
    finally:
        cursor.execute(f"DROP SECRET IF EXISTS IDENTIFIER('{secret_name}')")


def test_create_or_alter(api_integrations, secret, database, schema):
    ai_name = random_string(10, "test_api_integration_create_or_alter_")
    ai_def = ApiIntegration(
        name=ai_name,
        api_hook=GitHook(allow_any_secret=True),
        api_allowed_prefixes=["https://github.com"],
        enabled=True,
        api_blocked_prefixes=["https://github.com/snowflakedb/snowpy"],
        comment="created by test_create_or_alter",
    )
    ai = api_integrations[ai_name]
    ai.create_or_alter(ai_def)
    try:
        fetched_ai = ai.fetch()
        assert fetched_ai.name == ai_name.upper()
        assert isinstance(fetched_ai.api_hook, GitHook)

        fetched_ai.api_hook.allowed_authentication_secrets = [secret]
        fetched_ai.enabled = False
        fetched_ai.api_blocked_prefixes = None
        fetched_ai.comment = "altered by test_create_or_alter"

        ai.create_or_alter(fetched_ai)
        fetched_ai = ai.fetch()
        assert fetched_ai.name == ai_name.upper()
        assert isinstance(fetched_ai.api_hook, GitHook)
        assert fetched_ai.api_hook.allow_any_secret is False
        assert fetched_ai.api_hook.allowed_authentication_secrets == [
            f"{database.name.upper()}.{schema.name.upper()}.{secret.upper()}"
        ]
        assert (
            fetched_ai.api_hook.allowed_api_authentication_integrations
            == ai_def.api_hook.allowed_api_authentication_integrations
        )
        assert fetched_ai.api_allowed_prefixes == ai_def.api_allowed_prefixes
        assert fetched_ai.enabled is False
        # Simulated CoA doesn't support unsetting api_blocked_prefixes
        assert fetched_ai.api_blocked_prefixes == ai_def.api_blocked_prefixes
        assert fetched_ai.comment == "altered by test_create_or_alter"
    finally:
        ai.drop()



"""
Cortex Analyst API.

The Snowflake Cortex Analyst API is a REST API that allows end user to chat with their data leveraging semantic models to generate SQL queries.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, Field, StrictStr


class AnnotatedSqlQuery(BaseModel):
    """A model object representing the AnnotatedSqlQuery resource.

    Constructs an object of type AnnotatedSqlQuery with the provided properties.

    Parameters
    __________
    sql_text : str
        SQL query text
    database : str, optional
        Database associated with the SQL query
    var_schema : str, optional
        Schema associated with the SQL query
    corresponding_question : str, optional
        Natural language question that the SQL query answers
    additional_context : str, optional
        Free-form string for additional context about the query
    """

    sql_text: StrictStr

    database: Optional[StrictStr] = None

    var_schema: Optional[StrictStr] = Field(default=None, alias="schema")

    corresponding_question: Optional[StrictStr] = None

    additional_context: Optional[StrictStr] = None

    __properties = ["sql_text", "database", "schema", "corresponding_question", "additional_context"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> AnnotatedSqlQuery:
        """Create an instance of AnnotatedSqlQuery from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> AnnotatedSqlQuery:
        """Create an instance of AnnotatedSqlQuery from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return AnnotatedSqlQuery.model_validate(obj)

        _obj = AnnotatedSqlQuery.model_validate(
            {
                "sql_text": obj.get("sql_text"),
                "database": obj.get("database"),
                "var_schema": obj.get("schema"),
                "corresponding_question": obj.get("corresponding_question"),
                "additional_context": obj.get("additional_context"),
            }
        )

        return _obj


class AnnotatedSqlQueryModel:
    def __init__(
        self,
        sql_text: str,
        # optional properties
        database: Optional[str] = None,
        var_schema: Optional[str] = None,
        corresponding_question: Optional[str] = None,
        additional_context: Optional[str] = None,
    ):
        """A model object representing the AnnotatedSqlQuery resource.

        Constructs an object of type AnnotatedSqlQuery with the provided properties.

        Parameters
        __________
        sql_text : str
            SQL query text
        database : str, optional
            Database associated with the SQL query
        var_schema : str, optional
            Schema associated with the SQL query
        corresponding_question : str, optional
            Natural language question that the SQL query answers
        additional_context : str, optional
            Free-form string for additional context about the query
        """
        self.sql_text = sql_text
        self.database = database
        self.var_schema = var_schema
        self.corresponding_question = corresponding_question
        self.additional_context = additional_context

    __properties = ["sql_text", "database", "schema", "corresponding_question", "additional_context"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return AnnotatedSqlQuery(
            sql_text=self.sql_text,
            database=self.database,
            var_schema=self.var_schema,
            corresponding_question=self.corresponding_question,
            additional_context=self.additional_context,
        )

    @classmethod
    def _from_model(cls, model) -> AnnotatedSqlQueryModel:
        return AnnotatedSqlQueryModel(
            sql_text=model.sql_text,
            database=model.database,
            var_schema=model.var_schema,
            corresponding_question=model.corresponding_question,
            additional_context=model.additional_context,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a AnnotatedSqlQuery.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the AnnotatedSqlQuery object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> AnnotatedSqlQueryModel:
        """Create an instance of AnnotatedSqlQuery from a dict.

        This method constructs a AnnotatedSqlQuery object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        AnnotatedSqlQuery
            A AnnotatedSqlQuery object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(AnnotatedSqlQuery.from_dict(obj))


AnnotatedSqlQuery._model_class = AnnotatedSqlQueryModel



import json
import warnings

from unittest.mock import PropertyMock, patch

from snowflake.core._internal.client_info import ClientInfo


@patch("snowflake.core._internal.client_info.ClientInfo.client_version", new_callable=PropertyMock)
def test_client_info(mock_client_version, fake_root):
    mock_client_version.return_value = "1.0.0"

    # Test Case 1: PyCore entry is missing from client_support_version_info
    # It only has entries for other clients.
    fake_root._query_for_client_info.return_value = json.loads("""
    [
      {
        "clientId": "DOTNETDriver",
        "clientAppId": ".NET",
        "minimumSupportedVersion": "2.0.15",
        "minimumNearingEndOfSupportVersion": "2.0.17",
        "recommendedVersion": "4.1.0",
        "deprecatedVersions": [],
        "_customSupportedVersions_": []
    },
    {
        "clientId": "GO",
        "clientAppId": "Go",
        "minimumSupportedVersion": "1.6.12",
        "minimumNearingEndOfSupportVersion": "1.6.14",
        "recommendedVersion": "1.11.1",
        "deprecatedVersions": [],
        "_customSupportedVersions_": ["1.1.5"]
    }
  ]""")

    fake_root._initialize_client_info()

    # In this case, if an entry for PyCore isn't there, the
    # ClientInfo should be conservative and always say that the current
    # version is supported.
    client_info = fake_root._client_info

    assert client_info.client_version == "1.0.0"
    assert client_info.version_is_supported()
    assert not client_info.version_is_nearing_end_of_support()

    assert client_info.minimum_supported_version is None
    assert client_info.end_of_support_version is None
    assert client_info.recommended_version is None

    # Test Case 2: PyCore entry is present in client_support_version_info,
    # PyCore and the current version is newer than the recommended version
    fake_root._query_for_client_info.return_value = json.loads("""
    [
      {
        "clientId": "DOTNETDriver",
        "clientAppId": ".NET",
        "minimumSupportedVersion": "2.0.15",
        "minimumNearingEndOfSupportVersion": "2.0.17",
        "recommendedVersion": "4.1.0",
        "deprecatedVersions": [],
        "_customSupportedVersions_": []
    },
    {
        "clientId": "PyCore",
        "clientAppId": "PyCore",
        "minimumSupportedVersion": "0.7.0",
        "minimumNearingEndOfSupportVersion": "0.8.0",
        "recommendedVersion": "0.9.0",
        "deprecatedVersions": [],
        "_customSupportedVersions_": ["1.1.5"]
    }
  ]""")

    fake_root._initialize_client_info()
    client_info = fake_root._client_info

    # Information is present and the current version is supported
    assert client_info.client_version == "1.0.0"
    assert client_info.version_is_supported()
    assert not client_info.version_is_nearing_end_of_support()

    assert client_info.minimum_supported_version == "0.7.0"
    assert client_info.end_of_support_version == "0.8.0"
    assert client_info.recommended_version == "0.9.0"

    # Test Case 3: PyCore entry is present in client_support_version_info,
    # but the current version is too old.
    fake_root._query_for_client_info.return_value = json.loads("""
       [
       {
           "clientId": "PyCore",
           "clientAppId": "PyCore",
           "minimumSupportedVersion": "1.7.0",
           "minimumNearingEndOfSupportVersion": "1.8.0",
           "recommendedVersion": "1.9.0",
           "deprecatedVersions": [],
           "_customSupportedVersions_": ["1.1.5"]
       }
     ]""")

    # Verify that the unsupported warning is triggered
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")

        # This should trigger a warning.
        fake_root._initialize_client_info()
        assert len(w) == 1
        assert "not supported" in str(w[-1].message)

    client_info = fake_root._client_info

    # This version is not supported, which means it is also nearing end of support.
    assert client_info.client_version == "1.0.0"
    assert not client_info.version_is_supported()
    assert client_info.version_is_nearing_end_of_support()

    assert client_info.minimum_supported_version == "1.7.0"
    assert client_info.end_of_support_version == "1.8.0"
    assert client_info.recommended_version == "1.9.0"

    # Test Case 4: PyCore entry is present in client_support_version_info,
    # but the current version is lower than the nearing-end-of-support version.
    fake_root._query_for_client_info.return_value = json.loads("""
          [
          {
              "clientId": "PyCore",
              "clientAppId": "PyCore",
              "minimumSupportedVersion": "0.7.0",
              "minimumNearingEndOfSupportVersion": "1.8.0",
              "recommendedVersion": "1.9.0",
              "deprecatedVersions": [],
              "_customSupportedVersions_": ["1.1.5"]
          }
        ]""")

    # Verify that the nearing-end-of-life warning is triggered
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")

        # This should trigger a warning.
        fake_root._initialize_client_info()
        assert len(w) == 1
        assert "nearing the end of support" in str(w[-1].message)

    client_info = fake_root._client_info

    # This version is not supported, but it is nearing the end of support.
    assert client_info.client_version == "1.0.0"
    assert client_info.version_is_supported()
    assert client_info.version_is_nearing_end_of_support()

    assert client_info.minimum_supported_version == "0.7.0"
    assert client_info.end_of_support_version == "1.8.0"
    assert client_info.recommended_version == "1.9.0"

    # Test Case 5: No client support info is returned from the server
    # Things should default to being supported.
    fake_root._query_for_client_info.return_value = json.loads("""
    [
    ]""")

    fake_root._initialize_client_info()

    # In this case, ClientInfo should be conservative and always say that the current
    # version is supported.
    client_info = fake_root._client_info

    # Test the case where the client support version info for PyCore is not available;
    # In this case, ClientInfo should be conservative and always say that the current
    # version is supported.
    assert client_info.client_version == "1.0.0"
    assert client_info.version_is_supported()
    assert not client_info.version_is_nearing_end_of_support()

    assert client_info.minimum_supported_version is None
    assert client_info.end_of_support_version is None
    assert client_info.recommended_version is None

    # Test Case 6: PyCore entry is present in client_support_version_info,
    # but is incomplete (missing required fields).
    fake_root._query_for_client_info.return_value = json.loads("""
       [
       {
           "clientId": "PyCore",
           "clientAppId": "PyCore",
           "minimumSupportedVersion": "0.7.0",
           "deprecatedVersions": [],
           "_customSupportedVersions_": ["1.1.5"]
       }
     ]""")

    fake_root._initialize_client_info()
    client_info = fake_root._client_info

    # Since fields are missing, we should fall back to assuming the client is supported.
    assert client_info.client_version == "1.0.0"
    assert client_info.version_is_supported()
    assert not client_info.version_is_nearing_end_of_support()

    assert client_info.minimum_supported_version is None
    assert client_info.end_of_support_version is None
    assert client_info.recommended_version is None

    # Test Case 7: PyCore entry is present in client_support_version_info,
    # but has poorly formatted version numbers.
    fake_root._query_for_client_info.return_value = json.loads("""
      [
      {
          "clientId": "PyCore",
          "clientAppId": "PyCore",
          "minimumSupportedVersion": "0.7.0-dev",
          "minimumNearingEndOfSupportVersion": "0.f.0",
          "recommendedVersion": "0.9..0",
          "deprecatedVersions": [],
          "_customSupportedVersions_": ["1.1.5"]
      }
    ]""")

    fake_root._initialize_client_info()
    client_info = fake_root._client_info

    # Since fields are poorly formatted, we should fall back to assuming the client is supported.
    assert client_info.client_version == "1.0.0"
    assert client_info.version_is_supported()
    assert not client_info.version_is_nearing_end_of_support()

    assert client_info.minimum_supported_version is None
    assert client_info.end_of_support_version is None
    assert client_info.recommended_version is None

    # Test Case 8: Test for None as ClientInfo
    client_info = ClientInfo(None)

    # Since fields are poorly formatted, we should fall back to assuming the client is supported.
    assert client_info.client_version == "1.0.0"
    assert client_info.version_is_supported()
    assert not client_info.version_is_nearing_end_of_support()

    assert client_info.minimum_supported_version is None
    assert client_info.end_of_support_version is None
    assert client_info.recommended_version is None



"""
Snowflake Iceberg Table API.

The Snowflake Iceberg Table API is a REST API that you can use to access, update, and perform certain actions on Iceberg Table resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re

from datetime import datetime
from typing import Any, List, Optional

from pydantic import BaseModel, ConfigDict, Field, StrictBool, StrictInt, StrictStr, field_validator
from typing_extensions import Annotated

from snowflake.core.iceberg_table._generated.models.constraint import Constraint, ConstraintModel
from snowflake.core.iceberg_table._generated.models.iceberg_table_column import (
    IcebergTableColumn,
    IcebergTableColumnModel,
)


class IcebergTable(BaseModel):
    """A model object representing the IcebergTable resource.

    Constructs an object of type IcebergTable with the provided properties.

    Parameters
    __________
    name : str
        Name of the iceberg table
    comment : str, optional
        user comment associated to an object in the dictionary
    change_tracking : bool, optional
        True if change tracking is enabled, allowing streams and CHANGES to be used on the entity.
    max_data_extension_time_in_days : int, optional
        Maximum number of days to extend data retention beyond the retention period to prevent a stream becoming stale.
    external_volume : str, optional
        Name of an external volume that will be used for persisted Iceberg metadata and data files.
    data_retention_time_in_days : int, optional
        number of days to retain the old version of deleted/updated data
    catalog_sync : str, optional
        Name of the catalog integration to sync this table
    catalog : str, optional
        Name of the catalog integration to use for iceberg tables
    storage_serialization_policy : str, optional
        Storage serialization policy used for managed Iceberg table. This include encodings and compressions
    created_on : datetime, optional
        Date and time when the iceberg table was created — **Read-only:** *any user-provided value will be ignored.*
    database_name : str, optional
        Database in which the iceberg table is stored — **Read-only:** *any user-provided value will be ignored.*
    schema_name : str, optional
        Schema in which the iceberg table is stored — **Read-only:** *any user-provided value will be ignored.*
    owner : str, optional
        Role that owns the iceberg table — **Read-only:** *any user-provided value will be ignored.*
    owner_role_type : str, optional
        The type of role that owns the iceberg table — **Read-only:** *any user-provided value will be ignored.*
    iceberg_table_type : str, optional
        Type of Iceberg table. UNMANAGED if the table is not managed by Snowflake. NOT ICEBERG otherwise — **Read-only:** *any user-provided value will be ignored.*
    catalog_table_name : str, optional
        Name of the table as recognized by the catalog.
    catalog_namespace : str, optional
        Catalog namespace for the table. The namespace defined when the table was created. Otherwise, the default namespace associated with the catalog integration used by the table. If you’re syncing the table to Snowflake Open Catalog, the default is null.
    can_write_metadata : str, optional
        Signifies whether Snowflake can write metadata to the location specified by the file_path — **Read-only:** *any user-provided value will be ignored.*
    cluster_by : list[str], optional
        Specifies one or more columns or column expressions in the table as the clustering key.
    columns : list[IcebergTableColumn], optional

    base_location : str, optional
        The path to a directory where Snowflake can write data and metadata files for the table.
    replace_invalid_characters : bool, optional
        Specifies whether to replace invalid characters in the column names
    auto_refresh : bool, optional
        Specifies whether to automatically refresh the table metadata — **Read-only:** *any user-provided value will be ignored.*
    metadata_file_path : str, optional
        Specifies the relative path of the Iceberg metadata file to use for column definitions.
    constraints : list[Constraint], optional
    """

    name: Annotated[str, Field(strict=True)]

    comment: Optional[StrictStr] = None

    change_tracking: Optional[StrictBool] = None

    max_data_extension_time_in_days: Optional[StrictInt] = None

    external_volume: Optional[StrictStr] = None

    data_retention_time_in_days: Optional[StrictInt] = None

    catalog_sync: Optional[StrictStr] = None

    catalog: Optional[StrictStr] = None

    storage_serialization_policy: Optional[StrictStr] = None

    created_on: Optional[datetime] = None

    database_name: Optional[Annotated[str, Field(strict=True)]] = None

    schema_name: Optional[Annotated[str, Field(strict=True)]] = None

    owner: Optional[Annotated[str, Field(strict=True)]] = None

    owner_role_type: Optional[Annotated[str, Field(strict=True)]] = None

    iceberg_table_type: Optional[StrictStr] = None

    catalog_table_name: Optional[StrictStr] = None

    catalog_namespace: Optional[StrictStr] = None

    can_write_metadata: Optional[StrictStr] = None

    cluster_by: Optional[List[StrictStr]] = None

    columns: Optional[List[IcebergTableColumn]] = None

    base_location: Optional[StrictStr] = None

    replace_invalid_characters: Optional[StrictBool] = None

    auto_refresh: Optional[StrictBool] = None

    metadata_file_path: Optional[StrictStr] = None

    constraints: Optional[List[Constraint]] = None

    __properties = [
        "name",
        "comment",
        "change_tracking",
        "max_data_extension_time_in_days",
        "external_volume",
        "data_retention_time_in_days",
        "catalog_sync",
        "catalog",
        "storage_serialization_policy",
        "created_on",
        "database_name",
        "schema_name",
        "owner",
        "owner_role_type",
        "iceberg_table_type",
        "catalog_table_name",
        "catalog_namespace",
        "can_write_metadata",
        "cluster_by",
        "columns",
        "base_location",
        "replace_invalid_characters",
        "auto_refresh",
        "metadata_file_path",
        "constraints",
    ]

    @field_validator("name")
    def name_validate_regular_expression(cls, v):
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    @field_validator("storage_serialization_policy")
    def storage_serialization_policy_validate_enum(cls, v):
        if v is None:
            return v
        if v not in ("COMPATIBLE", "OPTIMIZED"):
            raise ValueError("must validate the enum values ('COMPATIBLE','OPTIMIZED')")
        return v

    @field_validator("database_name")
    def database_name_validate_regular_expression(cls, v):
        if v is None:
            return v
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    @field_validator("schema_name")
    def schema_name_validate_regular_expression(cls, v):
        if v is None:
            return v
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    @field_validator("owner")
    def owner_validate_regular_expression(cls, v):
        if v is None:
            return v
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    @field_validator("owner_role_type")
    def owner_role_type_validate_regular_expression(cls, v):
        if v is None:
            return v
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> IcebergTable:
        """Create an instance of IcebergTable from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update(
                {
                    "created_on",
                    "database_name",
                    "schema_name",
                    "owner",
                    "owner_role_type",
                    "iceberg_table_type",
                    "can_write_metadata",
                    "auto_refresh",
                }
            )

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        # override the default output from pydantic by calling `to_dict()` of each item in columns (list)
        _items = []
        if self.columns:
            for _item in self.columns:
                if _item:
                    _items.append(_item.to_dict())
            _dict["columns"] = _items

        # override the default output from pydantic by calling `to_dict()` of each item in constraints (list)
        _items = []
        if self.constraints:
            for _item in self.constraints:
                if _item:
                    _items.append(_item.to_dict())
            _dict["constraints"] = _items

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> IcebergTable:
        """Create an instance of IcebergTable from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return IcebergTable.model_validate(obj)

        _obj = IcebergTable.model_validate(
            {
                "name": obj.get("name"),
                "comment": obj.get("comment"),
                "change_tracking": obj.get("change_tracking"),
                "max_data_extension_time_in_days": obj.get("max_data_extension_time_in_days"),
                "external_volume": obj.get("external_volume"),
                "data_retention_time_in_days": obj.get("data_retention_time_in_days"),
                "catalog_sync": obj.get("catalog_sync"),
                "catalog": obj.get("catalog"),
                "storage_serialization_policy": obj.get("storage_serialization_policy"),
                "created_on": obj.get("created_on"),
                "database_name": obj.get("database_name"),
                "schema_name": obj.get("schema_name"),
                "owner": obj.get("owner"),
                "owner_role_type": obj.get("owner_role_type"),
                "iceberg_table_type": obj.get("iceberg_table_type"),
                "catalog_table_name": obj.get("catalog_table_name"),
                "catalog_namespace": obj.get("catalog_namespace"),
                "can_write_metadata": obj.get("can_write_metadata"),
                "cluster_by": obj.get("cluster_by"),
                "columns": [IcebergTableColumn.from_dict(_item) for _item in obj.get("columns")]
                if obj.get("columns") is not None
                else None,
                "base_location": obj.get("base_location"),
                "replace_invalid_characters": obj.get("replace_invalid_characters"),
                "auto_refresh": obj.get("auto_refresh"),
                "metadata_file_path": obj.get("metadata_file_path"),
                "constraints": [Constraint.from_dict(_item) for _item in obj.get("constraints")]
                if obj.get("constraints") is not None
                else None,
            }
        )

        return _obj


class IcebergTableModel:
    def __init__(
        self,
        name: str,
        # optional properties
        comment: Optional[str] = None,
        change_tracking: Optional[bool] = None,
        max_data_extension_time_in_days: Optional[int] = None,
        external_volume: Optional[str] = None,
        data_retention_time_in_days: Optional[int] = None,
        catalog_sync: Optional[str] = None,
        catalog: Optional[str] = None,
        storage_serialization_policy: Optional[str] = None,
        created_on: Optional[datetime] = None,
        database_name: Optional[str] = None,
        schema_name: Optional[str] = None,
        owner: Optional[str] = None,
        owner_role_type: Optional[str] = None,
        iceberg_table_type: Optional[str] = None,
        catalog_table_name: Optional[str] = None,
        catalog_namespace: Optional[str] = None,
        can_write_metadata: Optional[str] = None,
        cluster_by: Optional[list[str]] = None,
        columns: Optional[list[IcebergTableColumn]] = None,
        base_location: Optional[str] = None,
        replace_invalid_characters: Optional[bool] = None,
        auto_refresh: Optional[bool] = None,
        metadata_file_path: Optional[str] = None,
        constraints: Optional[list[Constraint]] = None,
    ):
        """A model object representing the IcebergTable resource.

        Constructs an object of type IcebergTable with the provided properties.

        Parameters
        __________
        name : str
            Name of the iceberg table
        comment : str, optional
            user comment associated to an object in the dictionary
        change_tracking : bool, optional
            True if change tracking is enabled, allowing streams and CHANGES to be used on the entity.
        max_data_extension_time_in_days : int, optional
            Maximum number of days to extend data retention beyond the retention period to prevent a stream becoming stale.
        external_volume : str, optional
            Name of an external volume that will be used for persisted Iceberg metadata and data files.
        data_retention_time_in_days : int, optional
            number of days to retain the old version of deleted/updated data
        catalog_sync : str, optional
            Name of the catalog integration to sync this table
        catalog : str, optional
            Name of the catalog integration to use for iceberg tables
        storage_serialization_policy : str, optional
            Storage serialization policy used for managed Iceberg table. This include encodings and compressions
        created_on : datetime, optional
            Date and time when the iceberg table was created.
        database_name : str, optional
            Database in which the iceberg table is stored
        schema_name : str, optional
            Schema in which the iceberg table is stored
        owner : str, optional
            Role that owns the iceberg table
        owner_role_type : str, optional
            The type of role that owns the iceberg table
        iceberg_table_type : str, optional
            Type of Iceberg table. UNMANAGED if the table is not managed by Snowflake. NOT ICEBERG otherwise.
        catalog_table_name : str, optional
            Name of the table as recognized by the catalog.
        catalog_namespace : str, optional
            Catalog namespace for the table. The namespace defined when the table was created. Otherwise, the default namespace associated with the catalog integration used by the table. If you’re syncing the table to Snowflake Open Catalog, the default is null.
        can_write_metadata : str, optional
            Signifies whether Snowflake can write metadata to the location specified by the file_path.
        cluster_by : list[str], optional
            Specifies one or more columns or column expressions in the table as the clustering key.
        columns : list[IcebergTableColumn], optional

        base_location : str, optional
            The path to a directory where Snowflake can write data and metadata files for the table.
        replace_invalid_characters : bool, optional
            Specifies whether to replace invalid characters in the column names
        auto_refresh : bool, optional
            Specifies whether to automatically refresh the table metadata
        metadata_file_path : str, optional
            Specifies the relative path of the Iceberg metadata file to use for column definitions.
        constraints : list[Constraint], optional
        """
        self.name = name
        self.comment = comment
        self.change_tracking = change_tracking
        self.max_data_extension_time_in_days = max_data_extension_time_in_days
        self.external_volume = external_volume
        self.data_retention_time_in_days = data_retention_time_in_days
        self.catalog_sync = catalog_sync
        self.catalog = catalog
        self.storage_serialization_policy = storage_serialization_policy
        self.created_on = created_on
        self.database_name = database_name
        self.schema_name = schema_name
        self.owner = owner
        self.owner_role_type = owner_role_type
        self.iceberg_table_type = iceberg_table_type
        self.catalog_table_name = catalog_table_name
        self.catalog_namespace = catalog_namespace
        self.can_write_metadata = can_write_metadata
        self.cluster_by = cluster_by
        self.columns = columns
        self.base_location = base_location
        self.replace_invalid_characters = replace_invalid_characters
        self.auto_refresh = auto_refresh
        self.metadata_file_path = metadata_file_path
        self.constraints = constraints

    __properties = [
        "name",
        "comment",
        "change_tracking",
        "max_data_extension_time_in_days",
        "external_volume",
        "data_retention_time_in_days",
        "catalog_sync",
        "catalog",
        "storage_serialization_policy",
        "created_on",
        "database_name",
        "schema_name",
        "owner",
        "owner_role_type",
        "iceberg_table_type",
        "catalog_table_name",
        "catalog_namespace",
        "can_write_metadata",
        "cluster_by",
        "columns",
        "base_location",
        "replace_invalid_characters",
        "auto_refresh",
        "metadata_file_path",
        "constraints",
    ]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return IcebergTable(
            name=self.name,
            comment=self.comment,
            change_tracking=self.change_tracking,
            max_data_extension_time_in_days=self.max_data_extension_time_in_days,
            external_volume=self.external_volume,
            data_retention_time_in_days=self.data_retention_time_in_days,
            catalog_sync=self.catalog_sync,
            catalog=self.catalog,
            storage_serialization_policy=self.storage_serialization_policy,
            created_on=self.created_on,
            database_name=self.database_name,
            schema_name=self.schema_name,
            owner=self.owner,
            owner_role_type=self.owner_role_type,
            iceberg_table_type=self.iceberg_table_type,
            catalog_table_name=self.catalog_table_name,
            catalog_namespace=self.catalog_namespace,
            can_write_metadata=self.can_write_metadata,
            cluster_by=self.cluster_by,
            columns=[x._to_model() for x in self.columns] if self.columns is not None else None,
            base_location=self.base_location,
            replace_invalid_characters=self.replace_invalid_characters,
            auto_refresh=self.auto_refresh,
            metadata_file_path=self.metadata_file_path,
            constraints=[x._to_model() for x in self.constraints] if self.constraints is not None else None,
        )

    @classmethod
    def _from_model(cls, model) -> IcebergTableModel:
        return IcebergTableModel(
            name=model.name,
            comment=model.comment,
            change_tracking=model.change_tracking,
            max_data_extension_time_in_days=model.max_data_extension_time_in_days,
            external_volume=model.external_volume,
            data_retention_time_in_days=model.data_retention_time_in_days,
            catalog_sync=model.catalog_sync,
            catalog=model.catalog,
            storage_serialization_policy=model.storage_serialization_policy,
            created_on=model.created_on,
            database_name=model.database_name,
            schema_name=model.schema_name,
            owner=model.owner,
            owner_role_type=model.owner_role_type,
            iceberg_table_type=model.iceberg_table_type,
            catalog_table_name=model.catalog_table_name,
            catalog_namespace=model.catalog_namespace,
            can_write_metadata=model.can_write_metadata,
            cluster_by=model.cluster_by,
            columns=[IcebergTableColumnModel._from_model(x) for x in model.columns] if model.columns else None,
            base_location=model.base_location,
            replace_invalid_characters=model.replace_invalid_characters,
            auto_refresh=model.auto_refresh,
            metadata_file_path=model.metadata_file_path,
            constraints=[ConstraintModel._from_model(x) for x in model.constraints] if model.constraints else None,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a IcebergTable.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the IcebergTable object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> IcebergTableModel:
        """Create an instance of IcebergTable from a dict.

        This method constructs a IcebergTable object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        IcebergTable
            A IcebergTable object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(IcebergTable.from_dict(obj))


IcebergTable._model_class = IcebergTableModel



"""
Cortex Inference API.

OpenAPI 3.0 specification for the Cortex REST API  # noqa: E501

The version of the OpenAPI document: 0.1.0
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, Dict, List, Optional

from pydantic import BaseModel, ConfigDict, StrictStr


class ToolResultsToolResults(BaseModel):
    """A model object representing the ToolResultsToolResults resource.

    Constructs an object of type ToolResultsToolResults with the provided properties.

    Parameters
    __________
    tool_use_id : str
        Unique identifier linking this result to its corresponding tool use request
    name : str
        Name of the tool that was executed. Must match a tool name from the tools array
    content : list[object]
        Array of content elements produced by the tool execution
    status : str, optional
        status indicators for the tool execution
    result : object, optional
        Additional result data in an arbitrary structure
    """

    tool_use_id: StrictStr

    name: StrictStr

    content: List[Any]

    status: Optional[StrictStr] = None

    result: Optional[Dict[str, Any]] = None

    __properties = ["tool_use_id", "name", "content", "status", "result"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> ToolResultsToolResults:
        """Create an instance of ToolResultsToolResults from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> ToolResultsToolResults:
        """Create an instance of ToolResultsToolResults from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return ToolResultsToolResults.model_validate(obj)

        _obj = ToolResultsToolResults.model_validate(
            {
                "tool_use_id": obj.get("tool_use_id"),
                "name": obj.get("name"),
                "content": obj.get("content"),
                "status": obj.get("status"),
                "result": obj.get("result"),
            }
        )

        return _obj


class ToolResultsToolResultsModel:
    def __init__(
        self,
        tool_use_id: str,
        name: str,
        content: list[object],
        # optional properties
        status: Optional[str] = None,
        result: Optional[object] = None,
    ):
        """A model object representing the ToolResultsToolResults resource.

        Constructs an object of type ToolResultsToolResults with the provided properties.

        Parameters
        __________
        tool_use_id : str
            Unique identifier linking this result to its corresponding tool use request
        name : str
            Name of the tool that was executed. Must match a tool name from the tools array
        content : list[object]
            Array of content elements produced by the tool execution
        status : str, optional
            status indicators for the tool execution
        result : object, optional
            Additional result data in an arbitrary structure
        """
        self.tool_use_id = tool_use_id
        self.name = name
        self.content = content
        self.status = status
        self.result = result

    __properties = ["tool_use_id", "name", "content", "status", "result"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return ToolResultsToolResults(
            tool_use_id=self.tool_use_id,
            name=self.name,
            content=self.content,
            status=self.status,
            result=self.result,
        )

    @classmethod
    def _from_model(cls, model) -> ToolResultsToolResultsModel:
        return ToolResultsToolResultsModel(
            tool_use_id=model.tool_use_id,
            name=model.name,
            content=model.content,
            status=model.status,
            result=model.result,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a ToolResultsToolResults.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the ToolResultsToolResults object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> ToolResultsToolResultsModel:
        """Create an instance of ToolResultsToolResults from a dict.

        This method constructs a ToolResultsToolResults object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        ToolResultsToolResults
            A ToolResultsToolResults object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(ToolResultsToolResults.from_dict(obj))


ToolResultsToolResults._model_class = ToolResultsToolResultsModel



"""
Snowflake User Defined Function API.

The Snowflake User Defined Function API is a REST API that you can use to access, update, and perform certain actions on User Defined Function resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, Optional

from pydantic import ConfigDict

from snowflake.core.user_defined_function._generated.models.function_language import (
    FunctionLanguage,
)


class JavaScriptFunction(FunctionLanguage):
    """A model object representing the JavaScriptFunction resource.

    Constructs an object of type JavaScriptFunction with the provided properties.

    Parameters
    __________

    called_on_null_input : bool, optional
        Decide if the function/procedure can receive null input

    is_volatile : bool, optional
         Specifies the behavior of the UDF when returning results. This Field is deprecated for Procedure.

                    If true, UDF might return different values for different rows, even for the same input.
                    This field is deprecated for Procedure.

                    If false, UDF assumes that the function, when called with the same inputs, will always return the same result.
                    This guarantee is not checked. Specifying IMMUTABLE for a UDF that returns different values for the same input will result in undefined behavior.
    """

    __properties = ["language", "called_on_null_input", "is_volatile"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> JavaScriptFunction:
        """Create an instance of JavaScriptFunction from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        _dict["language"] = FunctionLanguage.get_child_model_discriminator_value("JavaScriptFunction")

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> JavaScriptFunction:
        """Create an instance of JavaScriptFunction from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return JavaScriptFunction.model_validate(obj)

        _obj = JavaScriptFunction.model_validate(
            {
                "called_on_null_input": obj.get("called_on_null_input"),
                "is_volatile": obj.get("is_volatile"),
            }
        )

        return _obj


class JavaScriptFunctionModel(FunctionLanguage):
    def __init__(
        self,  # optional properties
        called_on_null_input: Optional[bool] = None,
        is_volatile: Optional[bool] = None,
    ):
        """A model object representing the JavaScriptFunction resource.

        Constructs an object of type JavaScriptFunction with the provided properties.

        Parameters
        __________

        called_on_null_input : bool, optional
            Decide if the function/procedure can receive null input

        is_volatile : bool, optional
             Specifies the behavior of the UDF when returning results. This Field is deprecated for Procedure.

            If true, UDF might return different values for different rows, even for the same input.
            This field is deprecated for Procedure.

            If false, UDF assumes that the function, when called with the same inputs, will always return the same result.
            This guarantee is not checked. Specifying IMMUTABLE for a UDF that returns different values for the same input will result in undefined behavior.
        """
        super().__init__(
            called_on_null_input=called_on_null_input,
            is_volatile=is_volatile,
        )

    __properties = ["language", "called_on_null_input", "is_volatile"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return JavaScriptFunction(
            called_on_null_input=self.called_on_null_input,
            is_volatile=self.is_volatile,
        )

    @classmethod
    def _from_model(cls, model) -> JavaScriptFunctionModel:
        return JavaScriptFunctionModel(
            called_on_null_input=model.called_on_null_input,
            is_volatile=model.is_volatile,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a JavaScriptFunction.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the JavaScriptFunction object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> JavaScriptFunctionModel:
        """Create an instance of JavaScriptFunction from a dict.

        This method constructs a JavaScriptFunction object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        JavaScriptFunction
            A JavaScriptFunction object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(JavaScriptFunction.from_dict(obj))


JavaScriptFunction._model_class = JavaScriptFunctionModel



from __future__ import absolute_import

# flake8: noqa

# import apis into api package
from snowflake.core.database_role._generated.api.database_role_api import DatabaseRoleApi

__all__ = [
    "DatabaseRoleApi",
]



import pytest as pytest

from snowflake.core._common import CreateMode
from snowflake.core.procedure import JavaScriptFunction, Procedure, ReturnDataType
from tests.utils import random_string


@pytest.mark.min_sf_ver("8.38.0")
def test_iter(procedures):
    procs = []

    try:
        procedure_name_prefix = random_string(10, "test_procedure_")
        for i in range(5):
            proc_name = f"{procedure_name_prefix}_AProc_{i}"

            procedure = Procedure(
                name=proc_name,
                arguments=[],
                return_type=ReturnDataType(datatype="REAL"),
                language_config=JavaScriptFunction(),
                body="""return 3.14""",
            )
            proc = procedures.create(procedure, mode=CreateMode.or_replace)
            procs.append(proc)

        for i in range(3):
            proc_name = f"{procedure_name_prefix}_BProc_{i}"

            procedure = Procedure(
                name=proc_name,
                arguments=[],
                return_type=ReturnDataType(datatype="REAL"),
                language_config=JavaScriptFunction(),
                body="""return 3.14""",
            )
            proc = procedures.create(procedure, mode=CreateMode.or_replace)
            procs.append(proc)

        assert len([proc.name for proc in procedures.iter()]) >= 8
        assert len([proc.name for proc in procedures.iter(like="%APROC%")]) == 5
        assert len([proc.name for proc in procedures.iter(like="%BPROC%")]) == 3

    finally:
        for proc in procs:
            proc.drop()



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#


import pytest

from snowflake.core.database import Database, DatabaseCollection, DatabaseResource

from ..utils import random_string


pytestmark = pytest.mark.usefixtures("backup_database_schema")


@pytest.mark.min_sf_ver("9.8.0")
def test_create_or_alter_database(databases, temp_db: DatabaseResource):
    db_def = temp_db.fetch()
    db_def.comment = "my new comment"
    db_def.data_retention_time_in_days = 0
    db_def.default_ddl_collation = "en_US-trim"
    db_def.log_level = "INFO"
    db_def.max_data_extension_time_in_days = 7
    db_def.suspend_task_after_num_failures = 1
    db_def.trace_level = "ALWAYS"
    db_def.user_task_managed_initial_warehouse_size = "SMALL"
    db_def.serverless_task_min_statement_size = "SMALL"
    db_def.serverless_task_max_statement_size = "MEDIUM"
    db_def.user_task_timeout_ms = 3600001
    temp_db.create_or_alter(db_def)
    new_db = databases[temp_db.name].fetch()
    assert new_db.name in (temp_db.name, temp_db.name.upper())
    assert new_db.comment == "my new comment"
    assert new_db.data_retention_time_in_days == 0
    assert new_db.default_ddl_collation == "en_US-trim"
    assert new_db.max_data_extension_time_in_days == 7
    assert new_db.suspend_task_after_num_failures == 1
    assert new_db.user_task_managed_initial_warehouse_size == "SMALL"
    assert new_db.serverless_task_min_statement_size == "SMALL"
    assert new_db.serverless_task_max_statement_size == "MEDIUM"
    assert new_db.user_task_timeout_ms == 3600001


def test_resist_multi_statement_sql_injection(databases: DatabaseCollection):
    new_db_name = random_string(3, "test_db_resist_multi_statement_sql_injection_")
    sql_injection_comment = "'comment for disguise'; select '1'"

    new_db = Database(name=new_db_name, comment=sql_injection_comment)

    db = databases.create(new_db)
    try:
        db.fetch()
    finally:
        db.drop()



import json

from unittest import mock
from urllib.parse import quote

import pytest

from snowflake.core import PollingOperation
from snowflake.core.function import FunctionResource, ServiceFunction
from snowflake.core.function._generated import TagAssignment, TagReference
from snowflake.core.tag import TagValue
from snowflake.core.version import __version__ as VERSION

from ...utils import BASE_URL, extra_params, mock_http_response


API_CLIENT_REQUEST = "snowflake.core._generated.api_client.ApiClient.request"
FUNCTION = ServiceFunction(name="my_fn", arguments=[], service="my_service", endpoint="", path="/path/to/myapp")


@pytest.fixture
def functions(schema):
    return schema.functions


@pytest.fixture
def function(functions):
    return functions["my_fn()"]


def test_create_function(fake_root, functions):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/functions")
    kwargs = extra_params(
        query_params=[],
        body={
            "name": "my_fn",
            "arguments": [],
            "returns": "TEXT",
            "service": "my_service",
            "endpoint": "",
            "path": "/path/to/myapp",
            "function_type": "service-function",
        },
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        fn_res = functions.create(FUNCTION)
        assert isinstance(fn_res, FunctionResource)
        assert fn_res.name_with_args == "my_fn()"
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = functions.create_async(FUNCTION)
        assert isinstance(op, PollingOperation)
        fn_res = op.result()
        assert fn_res.name_with_args == "my_fn()"
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_iter_function(fake_root, functions):
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/functions")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        functions.iter()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = functions.iter_async()
        assert isinstance(op, PollingOperation)
        it = op.result()
        assert list(it) == []
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_fetch_function(fake_root, function):
    from snowflake.core.function._generated.models import ServiceFunction as ServiceFunctionModel

    model = ServiceFunctionModel(name="my_fn", arguments=[], service="my_service", endpoint="", path="/path/to/myapp")
    args = (fake_root, "GET", BASE_URL + f"/databases/my_db/schemas/my_schema/functions/{quote('my_fn()')}")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(model.to_json())
        function.fetch()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(model.to_json())
        op = function.fetch_async()
        assert isinstance(op, PollingOperation)
        fn = op.result()
        assert fn.to_dict() == FUNCTION.to_dict()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_drop_function(fake_root, function):
    args = (
        fake_root,
        "DELETE",
        BASE_URL + f"/databases/my_db/schemas/my_schema/functions/{quote('my_fn()')}",
    )
    kwargs = extra_params(query_params=[])

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        function.drop()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = function.drop_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_execute_function(fake_root, function):
    from snowflake.core.function._generated.models import ServiceFunction as ServiceFunctionModel

    model = ServiceFunctionModel(name="my_fn", arguments=[], service="my_service", endpoint="", path="/path/to/myapp")
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/functions/my_fn:execute")
    kwargs = extra_params(
        headers={
            "Accept": "application/json",
            "User-Agent": "python_api/" + VERSION,
            "Content-Type": "application/json",
        },
        body=[],
    )
    fetch_response = mock_http_response(model.to_json())
    execute_response = mock_http_response(json.dumps({"result": 1}))

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.side_effect = [fetch_response, execute_response]
        function.execute()
    mocked_request.assert_called_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.side_effect = [fetch_response, execute_response]
        op = function.execute_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_with(*args, **kwargs)


def test_set_tags(fake_root, function, tag):
    args = (
        fake_root,
        "POST",
        BASE_URL + f"/databases/my_db/schemas/my_schema/functions/{quote('my_fn()')}:set-tags",
    )
    tags = {tag: TagValue(value="value")}
    kwargs = extra_params(
        body=[
            TagAssignment(
                tag_value=v.value, tag_name=k.name, tag_schema=k.schema.name, tag_database=k.database.name
            ).to_dict()
            for k, v in tags.items()
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        function.set_tags(tags)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = function.set_tags_async(tags)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_unset_tags(fake_root, function, tag):
    args = (
        fake_root,
        "POST",
        BASE_URL + f"/databases/my_db/schemas/my_schema/functions/{quote('my_fn()')}:unset-tags",
    )
    tag_resources = {tag}
    kwargs = extra_params(
        body=[
            TagReference(
                tag_name=tag_res.name, tag_schema=tag_res.schema.name, tag_database=tag_res.database.name
            ).to_dict()
            for tag_res in tag_resources
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        function.unset_tags(tag_resources)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = function.unset_tags_async(tag_resources)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_get_tags(fake_root, function):
    args = (
        fake_root,
        "GET",
        BASE_URL + f"/databases/my_db/schemas/my_schema/functions/{quote('my_fn()')}:get-tags",
    )
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        assert function.get_tags() == {}
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = function.get_tags_async()
        assert isinstance(op, PollingOperation)
        assert op.result() == {}
    mocked_request.assert_called_once_with(*args, **kwargs)
